# Standard library
from collections import OrderedDict

# Third-party
import astropy.units as u
import numpy as np

__all__ = ['pack_prior_samples', 'save_prior_samples']

# These units and the order are required for the likelihood code
_name_to_unit = OrderedDict()
_name_to_unit['P'] = u.day
_name_to_unit['M0'] = u.radian
_name_to_unit['e'] = u.one
_name_to_unit['omega'] = u.radian


def pack_prior_samples(samples, rv_unit):
    """
    Pack a dictionary of prior samples as Astropy Quantity
    objects into a single 2D array. The prior samples dictionary
    must contain keys for:

        - ``P``, period
        - ``M0``, phase at t=0
        - ``e``, eccentricity
        - ``omega``, argument of periastron
        - ``jitter``, velocity jitter (optional)

    Parameters
    ----------
    samples : dict
        A dictionary of prior samples as `~astropy.units.Quantity` objects.
    rv_unit : `~astropy.units.UnitBase`
        The radial velocity data unit.

    Returns
    -------
    arr_samples : `numpy.ndarray`
        An array of ``n`` prior samples with shape ``(n, 5)``. If jitter was not
        passed in, all jitter values will be automatically set to 0.
    units : list
        A list of `~astropy.units.UnitBase` objects specifying the units for
        each column.

    """

    arrs = []
    units = []
    for name, unit in _name_to_unit.items():
        if unit == u.one:
            arr = np.asarray(samples[name])
        else:
            arr = samples[name].to(unit).value
        arrs.append(arr)
        units.append(unit)

    if 'jitter' not in samples:
        jitter = np.zeros_like(arrs[0])

    else:
        jitter = samples['jitter'].to(rv_unit).value
    arrs.append(jitter)
    units.append(rv_unit)

    return np.vstack(arrs).T, units


def save_prior_samples(f, samples, rv_unit, ln_prior_probs=None):
    """
    Save a dictionary of Astropy Quantity prior samples to
    an HDF5 file in a format expected and used by
    `thejoker.sampler.TheJoker`. The prior samples dictionary
    must contain keys for:

        - ``P``, period
        - ``M0``, phase at pericenter
        - ``e``, eccentricity
        - ``omega``, argument of periastron
        - ``jitter``, velocity jitter (optional)

    Parameters
    ----------
    f : str, :class:`h5py.File`, :class:`h5py.Group`, :class:`h5py.DataSet`
        A string filename, or an instantiated `h5py` class.
    samples : dict
        A dictionary of prior samples as `~astropy.units.Quantity`
        objects.
    rv_unit : `~astropy.units.UnitBase`
        The radial velocity data unit.

    Returns
    -------
    units : list
        A list of `~astropy.units.UnitBase` objects specifying the
        units for each column.

    """

    packed_samples, units = pack_prior_samples(samples, rv_unit)

    if isinstance(f, str):
        import h5py
        with h5py.File(f, 'a') as g:
            g.attrs['units'] = np.array([str(x) for x in units]).astype('|S6')
            g['samples'] = packed_samples

            if ln_prior_probs is not None:
                g['ln_prior_probs'] = ln_prior_probs

    else:
        f.attrs['units'] = np.array([str(x) for x in units]).astype('|S6')
        f['samples'] = packed_samples

        if ln_prior_probs is not None:
            f['ln_prior_probs'] = ln_prior_probs

    return units



# Third-party
import astropy.units as u
import numpy as np

# Package
from ..likelihood import design_matrix, tensor_vector_scalar, marginal_ln_likelihood

from .helpers import FakeData


class TestLikelihood(object):

    # TODO: repeated code
    def truths_to_nlp(self, truths):
        # P, M0, ecc, omega
        P = truths['P'].to(u.day).value
        M0 = truths['M0'].to(u.radian).value
        ecc = truths['e']
        omega = truths['omega'].to(u.radian).value
        return np.array([P, M0, ecc, omega, 0.])

    def setup(self):
        d = FakeData()
        self.datasets = d.datasets
        self.params = d.params
        self.truths = d.truths

    def test_design_matrix(self):

        data = self.datasets['binary']
        nlp = self.truths_to_nlp(self.truths['binary'])
        A = design_matrix(nlp, data, self.params['binary'])
        assert A.shape == (len(data), 2) # K, v0
        assert np.allclose(A[:,1], 1)

        # TODO: triple disabled
        # nlp = self.truths_to_nlp(self.truths['triple'])
        # A = design_matrix(nlp, data, self.params['triple'])
        # assert A.shape == (len(data), 3) # K, v0, v1
        # assert np.allclose(A[:,1], 1)
        # assert np.allclose(A[:,2], data._t_bmjd)

    def test_tensor_vector_scalar(self):

        data = self.datasets['binary']
        nlp = self.truths_to_nlp(self.truths['binary'])
        A = design_matrix(nlp, data, self.params['binary'])
        ATCinvA, p, chi2 = tensor_vector_scalar(A, data.ivar.value,
                                                data.rv.value)
        true_p = [self.truths['binary']['K'].value,
                  self.truths['binary']['v0'].value]
        assert np.allclose(p, true_p, rtol=1e-2)

        # --

        # TODO: triple disabled
        # data = self.data['triple']
        # nlp = self.truths_to_nlp(self.truths['triple'])
        # A = design_matrix(nlp, data, self.params['triple'])
        # ATCinvA, p, chi2 = tensor_vector_scalar(A, data.ivar.value,
        #                                         data.rv.value)
        #
        # true_p = [self.truths['triple']['K'].value, self.fd.v0.value, self.fd.v1.value]
        # assert np.allclose(p, true_p, rtol=1e-2)

    def test_marginal_ln_likelihood_P(self):
        """
        Check that the true period is the maximum likelihood period
        """
        data = self.datasets['circ_binary']
        params = self.params['circ_binary']
        true_nlp = self.truths_to_nlp(self.truths['circ_binary'])
        # print('ln_like', marginal_ln_likelihood(true_nlp, data, params))

        vals = np.linspace(true_nlp[0]-1., true_nlp[0]+1, 4096)
        lls = np.zeros_like(vals)
        for i,val in enumerate(vals):
            nlp = true_nlp.copy()
            nlp[0] = val
            lls[i] = marginal_ln_likelihood(nlp, data, params)

        assert np.allclose(true_nlp[0], vals[lls.argmax()], rtol=1E-4)

    def test_marginal_ln_likelihood_P_samples(self):
        """
        This test is a little crazy. We generate samples in P, M0
        """

        data = self.datasets['circ_binary']
        params = self.params['circ_binary']
        true_nlp = self.truths_to_nlp(self.truths['circ_binary'])

        n_samples = 16384
        P = np.random.uniform(true_nlp[0] - 2., true_nlp[0] + 2., size=n_samples)
        M0 = np.random.uniform(0, 2*np.pi, size=n_samples)

        lls = np.zeros_like(P)
        n_neg = 0
        for i in range(n_samples):
            nlp = true_nlp.copy()
            nlp[0] = P[i]
            nlp[1] = M0[i]
            lls[i] = marginal_ln_likelihood(nlp, data, params)

            A = design_matrix(nlp, data, params)
            _,p,_ = tensor_vector_scalar(A, data.ivar.value, data.rv.value)
            if p[0] < 0:
                n_neg += 1

        # rejection sample using the marginal likelihood
        uu = np.random.uniform(size=n_samples)
        idx = uu < np.exp(lls - lls.max())
        print("{} good samples".format(idx.sum()))

        assert idx.sum() > 1
        assert np.std(P[idx]) < 1.

        # import matplotlib.pyplot as plt

        # plt.figure()
        # bins = np.linspace(true_nlp[0] - 2., true_nlp[0] + 2., 32)
        # plt.hist(P, bins=bins, alpha=0.25, normed=True)
        # plt.hist(P[idx], bins=bins, alpha=0.4, normed=True)
        # plt.axvline(true_nlp[0], color='r')

        # plt.figure()
        # bins = np.linspace(0, 2*np.pi, 32)
        # plt.hist(M0, bins=bins, alpha=0.25, normed=True)
        # plt.hist(M0[idx], bins=bins, alpha=0.4, normed=True)
        # plt.axvline(true_nlp[1], color='r')

        # plt.show()



# Third-party
import astropy.units as u
import numpy as np
import time

# Package
from ...data import RVData
from ..likelihood import marginal_ln_likelihood
from ..fast_likelihood import batch_marginal_ln_likelihood
from .. import JokerParams, TheJoker
from .helpers import FakeData


def test_against_py():
    joker_params = JokerParams(P_min=8*u.day, P_max=32768*u.day,
                               jitter=0*u.m/u.s,
                               linear_par_Vinv=np.diag(np.zeros(2)))
    joker = TheJoker(joker_params)

    t = np.random.uniform(0, 250, 16) + 56831.324
    t.sort()

    rv = np.cos(t)
    rv_err = np.random.uniform(0.1, 0.2, t.size)

    data = RVData(t=t, rv=rv*u.km/u.s, stddev=rv_err*u.km/u.s)

    samples = joker.sample_prior(size=16384)

    chunk = []
    for k in samples:
        chunk.append(np.array(samples[k]))

    chunk = np.ascontiguousarray(np.vstack(chunk).T)

    t0 = time.time()
    cy_ll = batch_marginal_ln_likelihood(chunk, data, joker_params)
    print("Cython:", time.time() - t0)

    t0 = time.time()
    n_chunk = len(chunk)
    py_ll = np.zeros(n_chunk)
    for i in range(n_chunk):
        try:
            py_ll[i] = marginal_ln_likelihood(chunk[i], data, joker_params)
        except Exception as e:
            py_ll[i] = np.nan
    print("Python:", time.time() - t0)

    assert np.allclose(np.array(cy_ll), py_ll)



# Third-party
import astropy.time as atime
import astropy.units as u
import numpy as np
import pytest

try:
    import matplotlib.pyplot as plt
    HAS_MPL = True
except:
    HAS_MPL = False

# Package
from ..data import RVData

def test_rvdata():

    # test various initializations
    t = np.random.uniform(55555., 56012., size=1024)
    rv = 100 * np.sin(0.5*t) * u.km/u.s
    ivar = 1 / (np.random.normal(0,5,size=1024)*u.km/u.s)**2
    RVData(t=t, rv=rv, ivar=ivar)

    t = atime.Time(t, format='mjd', scale='utc')
    RVData(t=t, rv=rv, ivar=ivar)

    with pytest.raises(TypeError):
        RVData(t=t, rv=rv.value, ivar=ivar)

    with pytest.raises(TypeError):
        RVData(t=t, rv=rv, ivar=ivar.value)

    # pass both
    with pytest.raises(ValueError):
        RVData(t=t, rv=rv, ivar=ivar, stddev=np.sqrt(1/ivar))

    # not velocity units
    with pytest.raises(u.UnitsError):
        RVData(t=t, rv=rv, ivar=ivar.value*u.km)

    # no error
    data = RVData(t=t, rv=rv)
    assert np.isnan(data.stddev.value).all()

    # shapes must be consistent
    with pytest.raises(ValueError):
        RVData(t=t[:-1], rv=rv, ivar=ivar)

    with pytest.raises(ValueError):
        RVData(t=t, rv=rv[:-1], ivar=ivar)

    with pytest.raises(ValueError):
        RVData(t=t, rv=rv, ivar=ivar[:-1])

    # check that copy works
    t = atime.Time(t, format='mjd', scale='utc')
    data1 = RVData(t=t, rv=rv, ivar=ivar)
    data2 = data1.copy()

    data1._t_bmjd += 1.5
    data1.rv *= 1.5
    data1.ivar *= 1.5
    assert np.all(data2._t_bmjd != data1._t_bmjd)
    assert np.all(data2.rv != data1.rv)
    assert np.all(data2.ivar != data1.ivar)

    # check slicing
    t = atime.Time(t, format='mjd', scale='utc')
    data1 = RVData(t=t, rv=rv, ivar=ivar)
    data2 = data1[:16]
    assert len(data2) == 16
    assert len(data2.t) == 16
    assert len(data2.rv) == 16
    assert len(data2.ivar) == 16

    # check filtering NaN's
    t = np.random.uniform(55555., 56012., size=128)
    rv = 100 * np.sin(0.5*t)
    rv[:16] = np.nan
    rv = rv * u.km/u.s
    ivar = 1 / (np.random.normal(0,5,size=t.size)*u.km/u.s)**2

    data = RVData(t=t, rv=rv, ivar=ivar)
    assert len(data) == (128-16)


@pytest.mark.skipif(not HAS_MPL, reason='matplotlib not installed')
def test_plotting():
    # check that plotting at least succeeds with allowed arguments
    t = np.random.uniform(55555., 56012., size=128)
    rv = 100 * np.sin(0.5*t) * u.km/u.s
    ivar = 1 / (np.random.normal(0,5,size=t.size)*u.km/u.s)**2
    data = RVData(t=t, rv=rv, ivar=ivar)

    data.plot()

    # style
    data.plot(color='r')

    # custom axis
    fig,ax = plt.subplots(1,1)
    data.plot(ax=plt.gca())

    # formatting
    data.plot(rv_unit=u.m/u.s)
    data.plot(rv_unit=u.m/u.s, time_format='jd')
    data.plot(rv_unit=u.m/u.s, time_format=lambda x: x.utc.mjd)

    # try with no errors
    data = RVData(t=t, rv=rv)
    data.plot()
    data.plot(ecolor='r')

    plt.close('all')



# Third-party
from astropy.time import Time
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import h5py
import numpy as np
import pytest

# Project
from ..samples import JokerSamples


def test_joker_samples(tmpdir):
    N = 100

    # Generate some fake samples
    samples = JokerSamples(P=np.random.random(size=N)*u.day)
    assert 'P' in samples
    assert len(samples) == N

    # Test slicing with a number
    s2 = samples[:10]
    assert len(s2) == 10

    # Invalid name
    with pytest.raises(ValueError):
        samples = JokerSamples(derp=15.)

    # Length conflicts with previous length
    samples = JokerSamples(P=np.random.random(size=N)*u.day)
    with pytest.raises(ValueError):
        samples['v0'] = np.random.random(size=10)*u.km/u.s

    # Write to HDF5 file
    samples = JokerSamples()
    samples['P'] = np.random.uniform(800, 1000, size=N)*u.day
    samples['M0'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['e'] = np.random.random(size=N)
    samples['omega'] = 2*np.pi*np.random.random(size=N)*u.radian

    fn = str(tmpdir / 'test.hdf5')
    with h5py.File(fn, 'w') as f:
        samples.to_hdf5(f)

    with h5py.File(fn, 'r') as f:
        samples2 = JokerSamples.from_hdf5(f)

    for k in samples.keys():
        assert quantity_allclose(samples[k], samples2[k])

    new_samples = samples[samples['P'].argmin()]
    assert quantity_allclose(new_samples['P'],
                             samples['P'][samples['P'].argmin()])

    # Check that t0 writes / gets loaded
    samples = JokerSamples(t0=Time('J2000'))
    samples['P'] = np.random.uniform(800, 1000, size=N)*u.day
    samples['M0'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['e'] = np.random.random(size=N)
    samples['omega'] = 2*np.pi*np.random.random(size=N)*u.radian

    fn = str(tmpdir / 'test.hdf5')
    with h5py.File(fn, 'w') as f:
        samples.to_hdf5(f)

    with h5py.File(fn, 'r') as f:
        samples2 = JokerSamples.from_hdf5(f)

    assert samples2.t0 is not None
    assert np.isclose(samples2.t0.mjd, Time('J2000').mjd)

    # Check that scalar samples are supported:
    samples = JokerSamples(t0=Time('J2000'))
    samples['P'] = np.random.uniform(800, 1000, size=N)*u.day
    samples['M0'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['e'] = np.random.random(size=N)
    samples['omega'] = 2*np.pi*np.random.random(size=N)*u.radian
    new_samples = samples[0]

    assert new_samples.shape == ()
    assert new_samples.size == 1

    # Check that polynomial trends work
    samples = JokerSamples(t0=Time('J2015.5'),
                           poly_trend=3)
    samples['P'] = np.random.uniform(800, 1000, size=N)*u.day
    samples['M0'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['e'] = np.random.random(size=N)
    samples['omega'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['K'] = 100 * np.random.normal(size=N) * u.km/u.s
    samples['v0'] = np.random.uniform(0, 10, size=N) * u.km/u.s
    samples['v1'] = np.random.uniform(0, 1, size=N) * u.km/u.s/u.day
    samples['v2'] = np.random.uniform(0, 1e-2, size=N) * u.km/u.s/u.day**2
    new_samples = samples[0]

    orb = samples.get_orbit(0)
    orb.radial_velocity(Time('J2015.6'))

    orb = new_samples.get_orbit(0)
    orb.radial_velocity(Time('J2015.6'))


def test_apply_methods():
    N = 100

    # Test that samples objects reduce properly
    samples = JokerSamples(t0=Time('J2000'))
    samples['P'] = np.random.uniform(800, 1000, size=N)*u.day
    samples['M0'] = 2*np.pi*np.random.random(size=N)*u.radian
    samples['e'] = np.random.random(size=N)
    samples['omega'] = 2*np.pi*np.random.random(size=N)*u.radian

    new_samples = samples.mean()
    assert quantity_allclose(new_samples['P'], np.mean(samples['P']))

    # try just executing others:
    new_samples = samples.median()
    new_samples = samples.std()



# Third-party
import astropy.units as u
import h5py
import numpy as np

# Package
from ...data import RVData
from ..io import pack_prior_samples, save_prior_samples


class TestIO(object):

    def setup(self):
        mjd = np.linspace(55555., 56555., 256)
        self.data = RVData(mjd, np.sin(mjd)*u.km/u.s,
                           stddev=0.1*np.random.uniform(size=mjd.size)*u.km/u.s)

        n = 128
        samples = dict()
        samples['P'] = np.random.uniform(size=n) * u.day
        samples['M0'] = np.random.uniform(size=n) * u.radian
        samples['omega'] = np.random.uniform(size=n) * u.radian
        samples['e'] = np.random.uniform(size=n) * u.one
        samples['jitter'] = np.random.uniform(size=n) * u.m/u.s
        self.samples = samples
        self.n = n

    def test_pack_prior_samples(self):
        samples = self.samples.copy()

        M,units = pack_prior_samples(samples, self.data.rv.unit)
        assert units[-1] == u.km/u.s
        assert M.shape == (self.n, 5)

        samples.pop('jitter')
        assert 'jitter' not in samples
        M,units = pack_prior_samples(samples, self.data.rv.unit)
        assert units[-1] == u.km/u.s
        assert M.shape == (self.n, 5)

    def test_save_prior_samples(self, tmpdir):

        path = str(tmpdir.join('io-test1.hdf5'))
        save_prior_samples(path, self.samples, self.data.rv.unit)
        with h5py.File(path, 'r') as f:
            assert f['samples'][:].shape == (self.n, 5)

        path = str(tmpdir.join('io-test2.hdf5'))
        with h5py.File(path, 'w') as f:
            save_prior_samples(f, self.samples, self.data.rv.unit)

        with h5py.File(path, 'r') as f:
            assert f['samples'][:].shape == (self.n, 5)



# Standard library
from collections import OrderedDict
import copy
import warnings

# Third-party
import astropy.units as u
from astropy.time import Time
import numpy as np
from twobody import KeplerOrbit, PolynomialRVTrend

# Package
from ..utils import quantity_to_hdf5, quantity_from_hdf5

__all__ = ['JokerSamples']


class JokerSamples(OrderedDict):
    _valid_keys = ['P', 'M0', 'e', 'omega', 'jitter', 'K']

    def __init__(self, t0=None, poly_trend=1, **kwargs):
        """A dictionary-like object for storing posterior samples from
        The Joker, with some extra functionality.

        Parameters
        ----------
        t0 : `astropy.time.Time`, numeric (optional)
            The reference time for the orbital parameters.
        poly_trend : int, optional
            If specified, sample over a polynomial velocity trend with the
            specified number of coefficients. For example, ``poly_trend=3`` will
            sample over parameters of a long-term quadratic velocity trend.
            Default is 1, just a constant velocity shift.
        **kwargs
            These are the orbital element names.
        """

        # initialize empty dictionary
        super(JokerSamples, self).__init__()
        self._setup(poly_trend, t0)

        for key, val in kwargs.items():
            self[key] = val # calls __setitem__ below

    def _setup(self, poly_trend, t0):
        # reference time
        self.t0 = t0

        self._size = None
        self._shape = None

        self._cache = dict()

        self.poly_trend = int(poly_trend)
        self._trend_names = ['v{0}'.format(i)
                             for i in range(self.poly_trend)]
        for name in self._trend_names:
            if name not in self._valid_keys:
                self._valid_keys.append(name)

    def __getstate__(self):
        return (self.poly_trend, self.t0, dict(self))

    def __setstate__(self, state):
        poly_trend, t0, data = state
        self._setup(poly_trend, t0)
        self.update(data)

    def __reduce__(self):
        return (JokerSamples, (), self.__getstate__())

    def _validate_key(self, key):
        if key not in self._valid_keys:
            raise ValueError("Invalid key '{0}'.".format(key))

    def _validate_val(self, val):
        val = u.Quantity(val)
        if self._shape is not None and val.shape != self.shape:
            raise ValueError("Shape of new samples must match those already "
                             "stored! ({0}, expected {1})"
                             .format(len(val), self.shape))

        return val

    def __getitem__(self, slc):
        if isinstance(slc, str):
            return super(JokerSamples, self).__getitem__(slc)

        else:
            new = copy.copy(self)
            new._size = None # reset number of samples
            new._shape = None # reset number of samples

            for k in self.keys():
                new[k] = self[k][slc]

            return new

    def __setitem__(self, key, val):
        self._validate_key(key)
        val = self._validate_val(val)

        if self._shape is None:
            self._shape = val.shape
            self._size = val.size

        super(JokerSamples, self).__setitem__(key, val)

    @property
    def n_samples(self):
        warnings.warn(".n_samples is deprecated in favor of .size",
                      DeprecationWarning)
        return self.size

    @property
    def size(self):
        if self._size is None:
            raise ValueError("No samples stored!")
        return self._size

    @property
    def shape(self):
        if self._shape is None:
            raise ValueError("No samples stored!")
        return self._shape

    def __len__(self):
        return self.n_samples

    def __str__(self):
        return ("<JokerSamples in [{0}], {1} samples>"
                .format(','.join(self.keys()), len(self)))

    @classmethod
    def from_hdf5(cls, f, n=None, **kwargs):
        """
        Parameters
        ----------
        f : :class:`h5py.File`, :class:`h5py.Group`
        n : int (optional)
            The number of samples to load.
        **kwargs
            All other keyword arguments are passed to the class initializer.
        """

        if 't0_bmjd' in f.attrs:
            # Read the reference time:
            t0 = Time(f.attrs['t0_bmjd'], format='mjd', scale='tcb')
        else:
            t0 = None

        samples = cls(t0=t0, **kwargs)
        for key in cls._valid_keys:
            if key in f:
                samples[key] = quantity_from_hdf5(f, key, n=n)

        return samples

    def to_hdf5(self, f):
        """
        Parameters
        ----------
        f : :class:`h5py.File`, :class:`h5py.Group`
        """

        for key in self.keys():
            quantity_to_hdf5(f, key, self[key])

        if self.t0 is not None:
            f.attrs['t0_bmjd'] = self.t0.tcb.mjd

    ##########################################################################
    # Interaction with TwoBody

    def get_orbit(self, index=None, **kwargs):
        """Get a `twobody.KeplerOrbit` object for the samples at the specified
        index.

        Parameters
        ----------
        index : int (optional)
            The index of the samples to turn into a `twobody.KeplerOrbit`
            instance. If the samples object is scalar, no index is necessary.
        **kwargs
            Other keyword arguments are passed to the `twobody.KeplerOrbit`
            initializer. For example, you can specify the inclination by passing
            ``i=...`, or  longitude of the ascending node by passing
            ``Omega=...``.

        Returns
        -------
        orbit : `twobody.KeplerOrbit`
            The samples converted to an orbit object. The barycenter position
            and distance are set to arbitrary values.
        """
        if 'orbit' not in self._cache:
            self._cache['orbit'] = KeplerOrbit(P=1*u.yr, e=0., omega=0*u.deg,
                                               Omega=0*u.deg, i=90*u.deg,
                                               a=1*u.au, t0=self.t0)

        # all of this to avoid the __init__ of KeplerOrbit / KeplerElements
        orbit = copy.copy(self._cache['orbit'])

        P = self['P']
        e = self['e']
        K = self['K']
        omega = self['omega']
        M0 = self['M0']
        a = kwargs.pop('a', P * K / (2*np.pi) * np.sqrt(1 - e**2))

        if len(self) == 1 and len(self.shape) == 0:
            if index > 0:
                raise ValueError('Samples are scalar-valued!')

            trend_coeffs = [self[x] for x in self._trend_names]

        else:
            P = P[index]
            e = e[index]
            a = a[index]
            omega = omega[index]
            M0 = M0[index]
            trend_coeffs = [self[x][index] for x in self._trend_names]

        orbit.elements._P = P
        orbit.elements._e = e * u.dimensionless_unscaled
        orbit.elements._a = a
        orbit.elements._omega = omega
        orbit.elements._M0 = M0
        orbit.elements._Omega = kwargs.pop('Omega', 0*u.deg)
        orbit.elements._i = kwargs.pop('i', 90*u.deg)
        orbit._vtrend = PolynomialRVTrend(trend_coeffs, t0=self.t0)
        orbit._barycenter = kwargs.pop('barycenter', None)

        if kwargs:
            raise ValueError("Unrecognized arguments {0}"
                             .format(', '.join(list(kwargs.keys()))))

        return orbit

    @property
    def orbits(self):
        """A generator that successively returns `twobody.KeplerOrbit` objects
        for each sample. See docstring for `thejoker.JokerSamples.get_orbit` for
        more information.

        """
        for i in range(len(self)):
            yield self.get_orbit(i)

    # Numpy reduce function
    def _apply(self, func):
        cls = self.__class__

        kw = dict()
        for k in self.keys():
            kw[k] = func(self[k])

        kw['t0'] = self.t0
        kw['poly_trend'] = self.poly_trend
        return cls(**kw)

    def mean(self):
        """Return a new scalar object by taking the mean across all samples"""
        return self._apply(np.mean)

    def median(self):
        """Return a new scalar object by taking the medin across all samples"""
        return self._apply(np.mean)

    def std(self):
        """Return a new scalar object by taking the medin across all samples"""
        return self._apply(np.std)



# Third-party
from numpy import log, pi
from scipy.special import loggamma

__all__ = ['beta_logpdf']


def beta_logpdf(x, a, b):
    denom = loggamma(a) + loggamma(b)
    res = loggamma(a+b) + (a-1)*log(x) + (b-1)*log(1-x) - denom
    return res.real


def norm_logpdf(x, mu, sig):
    return -0.5 * (((x-mu) / sig)**2 + log(2*pi*sig**2))



# Standard library
from collections import OrderedDict

# Third-party
from astropy.time import Time
import astropy.units as u
import numpy as np
from twobody import KeplerOrbit

# Package
from ...data import RVData
from ..params import JokerParams


class FakeData(object):

    def __init__(self, seed=42):
        rnd = np.random.RandomState(seed=seed)

        EPOCH = Time('J2000') + rnd.uniform(0., 40) * u.day

        self.datasets = OrderedDict()
        self.params = OrderedDict()
        self.truths = OrderedDict()

        P = rnd.uniform(40, 80) * u.day

        t = Time('J2000') + rnd.uniform(0, 300., 8) * u.day

        ######################################################################
        # Binary with random parameters

        truth = dict()
        truth['P'] = P
        truth['M0'] = rnd.uniform(0., 2*np.pi) * u.radian
        truth['omega'] = rnd.uniform(0., 2*np.pi) * u.radian
        truth['e'] = rnd.uniform()
        truth['K'] = rnd.uniform(5, 15) * u.km/u.s
        truth['v0'] = rnd.uniform(-100, 100) * u.km/u.s

        orbit = KeplerOrbit(P=truth['P'], e=truth['e'], omega=truth['omega'],
                            M0=truth['M0'], t0=EPOCH,
                            i=90*u.deg, Omega=0*u.deg) # these don't matter

        rv = truth['K'] * orbit.unscaled_radial_velocity(t) + truth['v0']
        err = np.full_like(rv.value, 0.01) * u.km/u.s
        data = RVData(t, rv, stddev=err, t0=EPOCH)
        self.datasets['binary'] = data
        self.params['binary'] = JokerParams(P_min=8*u.day, P_max=1024*u.day,
                                            jitter=1*u.km/u.s)
        self.truths['binary'] = truth.copy()

        # Binary with samplin over jitter
        self.datasets['binary_jitter'] = data
        self.params['binary_jitter'] = JokerParams(P_min=8*u.day,
                                                   P_max=1024*u.day,
                                                   jitter=(9., 2),
                                                   jitter_unit=u.m/u.s)
        self.truths['binary_jitter'] = truth.copy()
        self.truths['binary_jitter']['jitter'] = 0. * u.km/u.s

        ######################################################################
        # Binary on circular orbit

        truth = dict()
        truth['P'] = P
        truth['K'] = rnd.uniform(5, 15) * u.km/u.s
        truth['M0'] = rnd.uniform(0., 2*np.pi) * u.radian
        truth['omega'] = 0*u.radian
        truth['e'] = 0.
        truth['v0'] = rnd.uniform(-100, 100) * u.km/u.s

        orbit = KeplerOrbit(P=truth['P'], e=truth['e'], omega=truth['omega'],
                            M0=truth['M0'], t0=EPOCH,
                            i=90*u.deg, Omega=0*u.deg)

        rv = truth['K'] * orbit.unscaled_radial_velocity(t) + truth['v0']
        err = np.full_like(rv.value, 0.1) * u.km/u.s
        data = RVData(t, rv, stddev=err, t0=EPOCH)
        self.datasets['circ_binary'] = data
        self.params['circ_binary'] = JokerParams(P_min=8*u.day,
                                                 P_max=1024*u.day)
        self.truths['circ_binary'] = truth.copy()

        ######################################################################
        # hierarchical triple - long term velocity trend
        # TODO: no longer works with TwoBody changes
        # self.v1 = rnd.uniform(-1, 1) * u.km/u.s/u.day
        # orbit = RVOrbit(**truth)
        # rv = orbit.generate_rv_curve(mjd) + self.v0 + self.v1*(mjd-mjd.min())*u.day
        # err = np.full_like(rv.value, 0.01) * u.km/u.s
        # data = RVData(mjd, rv, stddev=err, t_offset=mjd.min())
        # self.data['triple'] = data
        # self.joker_params['triple'] = JokerParams(P_min=8*u.day,
        #                                           P_max=1024*u.day)
        # self.truths['triple'] = truth.copy()
        # self.truths['triple']['M0'] = self.truths['triple']['M0'] - ((2*np.pi*data.t_offset/P.value))*u.radian



# Third-party
import astropy.units as u
import pytest

# Project
from ..params import JokerParams


def test_init():

    pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day)
    assert pars.jitter.value == 0

    # test invalid input
    # with pytest.raises(TypeError):
    #     pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day, trend="derp")

    with pytest.raises(ValueError):
        pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day, jitter="derp")

    with pytest.raises(ValueError):
        pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day,
                           jitter="derp", jitter_unit="cat")

    with pytest.raises(TypeError):
        pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day,
                           jitter=(0.5, 1.), jitter_unit="cat")

    with pytest.raises(ValueError):
        pars = JokerParams(P_min=8.*u.day, P_max=8192*u.day,
                           jitter=(0.1, 5., 1.))



"""
NOTE: this is only used for testing the cython / c implementation.
"""

# Third-party
import numpy as np
from twobody.wrap import cy_rv_from_elements

# Package
from ..log import log as logger

__all__ = ['get_ivar', 'design_matrix',
           'tensor_vector_scalar', 'marginal_ln_likelihood']


def get_ivar(data, s):
    """Return a copy of the inverse variance array with jitter included.

    This is safe for zero'd out inverse variances.

    Parameters
    ----------
    data : `~thejoker.data.RVData`
    s : numeric
        Jitter in the same units as the RV data.

    """
    return data.ivar.value / (1 + s**2 * data.ivar.value)


def design_matrix(nonlinear_p, data, joker_params):
    """

    Parameters
    ----------
    nonlinear_p : array_like
        Array of non-linear parameter values. For the default case,
        these are P (period, day), M0 (phase at pericenter, rad),
        ecc (eccentricity), omega (argument of perihelion, rad).
        May also contain log(jitter^2) as the last index.
    data : `~thejoker.data.RVData`
        The observations.
    joker_params : `~thejoker.sampler.params.JokerParams`
        The specification of parameters to infer with The Joker.

    Returns
    -------
    A : `numpy.ndarray`
        The design matrix with shape ``(n_times, n_params)``.

    """
    P, M0, ecc, omega = nonlinear_p[:4] # we don't need the jitter here

    t = data._t_bmjd
    t0 = data._t0_bmjd
    zdot = cy_rv_from_elements(t, P, 1., ecc, omega, M0, t0,
                               joker_params.anomaly_tol,
                               joker_params.anomaly_maxiter)

    A1 = np.vander(t - t0, N=joker_params.poly_trend, increasing=True)
    A = np.hstack((zdot[:, None], A1))

    return A


def tensor_vector_scalar(A, ivar, y):
    """
    Internal function used to construct linear algebra objects
    used to compute the marginal log-likelihood.

    Parameters
    ----------
    A : `~numpy.ndarray`
        Design matrix.
    ivar : `~numpy.ndarray`
        Inverse-variance matrix.
    y : `~numpy.ndarray`
        Data (in this case, radial velocities).

    Returns
    -------
    ATCinvA : `numpy.ndarray`
        Value of A^T C^-1 A -- inverse of the covariance matrix
        of the linear parameters.
    p : `numpy.ndarray`
        Optimal values of linear parameters.
    chi2 : float
        Chi-squared value.

    Notes
    -----
    The linear parameter vector returned here (``p``) may have a negative
    velocity semi-amplitude. I don't think there is anything we can do
    about this if we want to preserve the simple linear algebra below and
    it means that optimizing over the marginal likelihood below won't
    work -- in the argument of periastron, there will be two peaks of
    similar height, and so the highest marginal likelihood period might be
    shifted from the truth.

    """
    ATCinv = (A.T * ivar[None])
    ATCinvA = ATCinv.dot(A)

    # Note: this is unstable! if cond num is high, could do:
    # p,*_ = np.linalg.lstsq(A, y)
    p = np.linalg.solve(ATCinvA, ATCinv.dot(y))
    dy = A.dot(p) - y

    chi2 = np.sum(dy**2 * ivar) # don't need log term for the jitter b.c. in likelihood below

    return ATCinvA, p, chi2


def marginal_ln_likelihood(nonlinear_p, data, joker_params, tvsi=None):
    """
    Internal function used to compute the likelihood marginalized
    over the linear parameters.

    This returns Eq. 11 arxiv:1610.07602v2

    Parameters
    ----------
    nonlinear_p : array_like
        Array of non-linear parameter values. For the default case,
        these are P (period, day), M0 (phase at pericenter, rad),
        ecc (eccentricity), omega (argument of perihelion, rad).
        May also contain jitter as the last index.
    data : `~thejoker.data.RVData`
        The observations.
    joker_params : `~thejoker.sampler.params.JokerParams`
        The specificationof parameters to infer with The Joker.
    tvsi : iterable (optional)
        Optionally pass in the tensor, vector, scalar, ivar values so they
        aren't re-computed.

    Returns
    -------
    marg_ln_like : `numpy.ndarray`
        Marginal log-likelihood values.

    """
    if tvsi is None:
        A = design_matrix(nonlinear_p, data, joker_params)

        # jitter must be in same units as the data RV's / ivar!
        s = nonlinear_p[4]
        ivar = get_ivar(data, s)
        ATCinvA, p, chi2 = tensor_vector_scalar(A, ivar, data.rv.value)

    else:
        ATCinvA, p, chi2, ivar = tvsi

    # This is -logdet(2πC_j)
    sign, logdet = np.linalg.slogdet(ATCinvA / (2*np.pi))
    if not np.all(sign == 1.):
        logger.debug('logdet sign < 0')
        return np.nan

    logdet += np.sum(np.log(ivar / (2*np.pi)))

    return 0.5*logdet - 0.5*np.atleast_1d(chi2)



# Standard library
import sys
import tempfile
import time

# Third-party
import astropy.units as u
from astropy.utils.misc import isiterable
import h5py
import numpy as np

# Project
from ..log import log as logger
from ..data import RVData
from ..stats import beta_logpdf, norm_logpdf
from .params import JokerParams
from .multiproc_helpers import (get_good_sample_indices, compute_likelihoods,
                                sample_indices_to_full_samples)
from .io import save_prior_samples
from .samples import JokerSamples
from .mcmc import TheJokerMCMCModel

__all__ = ['TheJoker']


class TheJoker:
    """A custom Monte-Carlo sampler for two-body systems.

    Parameters
    ----------
    params : `~thejoker.sampler.params.JokerParams`
        Object specifying hyper-parameters for The Joker.
    pool : ``schwimmbad.BasePool`` (optional)
        A processing pool (default is a ``schwimmbad.SerialPool`` instance).
    random_state : `numpy.random.RandomState` (optional)
        A ``RandomState`` instance to serve as a parent for the random
        number generators. See the :ref:`random numbers <random-numbers>` page
        for more information.
    n_batches : int (optional)
        When using multiprocessing to split the likelihood evaluations, this
        sets the number of batches to split the work into. Defaults to
        ``pool.size``, meaning equal work to each worker. For very large prior
        sample caches, you may need to set this to a larger number (e.g.,
        ``100*pool.size``) to avoid memory issues.
    tempfile_path : str (optional)
        Path to create temporary files needed for executing the sampler.
        Defaults to whatever ``tempfile`` thinks is a good temporary directory.
    """

    def __init__(self, params, pool=None, random_state=None, n_batches=None,
                 tempfile_path=None):

        # set the processing pool
        if pool is None:
            import schwimmbad
            pool = schwimmbad.SerialPool()

        elif not hasattr(pool, 'map') or not hasattr(pool, 'close'):
            raise TypeError("Input pool object must have .map() and .close() "
                            "methods. We recommend using `schwimmbad` pools.")

        self.pool = pool

        # Set the parent random state - child processes get different states
        # based on the parent
        if random_state is None:
            self._rnd_passed = False
            random_state = np.random.RandomState()

        elif not isinstance(random_state, np.random.RandomState):
            raise TypeError("Random state object must be a numpy RandomState "
                            "instance, not '{0}'".format(type(random_state)))

        else:
            self._rnd_passed = True

        self.random_state = random_state

        # check if a JokerParams instance was passed in to specify the state
        if not isinstance(params, JokerParams):
            raise TypeError("Parameter specification must be a JokerParams "
                            "instance, not a '{0}'".format(type(params)))
        self.params = params

        self.n_batches = n_batches
        self.tempfile_path = tempfile_path

    def sample_prior(self, size=1, return_logprobs=False):
        """Generate samples from the prior. Logarithmic in period, uniform in
        phase and argument of pericenter, Beta distribution in eccentricity.

        Parameters
        ----------
        size : int
            Number of samples to generate.
        return_logprobs : bool (optional)
            If ``True``, will also return the log-value of the prior at each
            sample.

        Returns
        -------
        samples : `~thejoker.sampler.samples.JokerSamples`
            Keys: `['P', 'M0', 'e', 'omega']`, each as
            `astropy.units.Quantity` objects (i.e. with units).

        TODO
        ----
        - All prior distributions are fixed. These should be customizable.
        """
        rnd = self.random_state

        # Create an empty, dictionary-like 'samples' object to fill
        samples = JokerSamples()

        # sample from priors in nonlinear parameters
        a, b = (np.log(self.params.P_min.to(u.day).value),
                np.log(self.params.P_max.to(u.day).value))
        samples['P'] = np.exp(rnd.uniform(a, b, size=size)) * u.day

        samples['M0'] = rnd.uniform(0, 2 * np.pi, size=size) * u.radian

        # MAGIC NUMBERS below: Kipping et al. 2013 (MNRAS 434 L51)
        samples['e'] = rnd.beta(a=0.867, b=3.03, size=size) * u.one

        samples['omega'] = rnd.uniform(0, 2 * np.pi, size=size) * u.radian

        # Store the value of the prior at each prior sample
        # TODO: should we store the value for each parameter independently?
        if return_logprobs:
            ln_prior_val = np.zeros(size)

            # P
            ln_prior_val += -np.log(b - a) - np.log(samples['P'].value)

            # M0
            ln_prior_val += -np.log(2 * np.pi)

            # e - MAGIC NUMBERS below: Kipping et al. 2013 (MNRAS 434 L51)
            ln_prior_val += beta_logpdf(samples['e'].value, 0.867, 3.03)

            # omega
            ln_prior_val += -np.log(2 * np.pi)

        if not self.params._fixed_jitter:
            # Gaussian prior in log(s^2)
            log_s2 = rnd.normal(*self.params.jitter, size=size)
            samples['jitter'] = np.sqrt(
                np.exp(log_s2)) * self.params._jitter_unit

            if return_logprobs:
                Jac = np.log(2 / samples['jitter'].value)  # Jacobian
                ln_prior_val += norm_logpdf(log_s2,
                                            self.params.jitter[0],
                                            self.params.jitter[1]) + Jac

        else:
            samples['jitter'] = np.ones(size) * self.params.jitter

        if return_logprobs:
            return samples, ln_prior_val
        else:
            return samples

    def _unpack_full_samples(self, result, prior_units, return_logprobs,
                             t0=None):
        """Unpack an array of The Joker samples into a dictionary-like object of
        Astropy Quantity objects (with units). This is meant to be used
        internally.

        Parameters
        ----------
        result : tuple
            A tuple of output directly from _rejection_sample_from_cache().
            Depending on the value of ``return_logprobs``, this is either just
            the sample values as a 2D array, or a lenth 3 tuple with the 2D
            samples array, prior values, and likelihood values.
        prior_units : list
            List of units for the prior samples.
        return_logprobs : bool
            Are we also returning the log prior values?
        t0 : `~astropy.time.Time` (optional)
            Passed to `thejoker.JokerSamples`.

        Returns
        -------
        samples : `~thejoker.sampler.samples.JokerSamples`

        """

        if return_logprobs:
            samples_arr, ln_prior, ln_like = result

        else:
            samples_arr = result

        n, n_params = samples_arr.shape

        samples = JokerSamples(t0=t0, poly_trend=self.params.poly_trend)

        # TODO: need to keep track of this elsewhere...
        nonlin_params = ['P', 'M0', 'e', 'omega', 'jitter']
        for k, key in enumerate(nonlin_params):
            samples[key] = samples_arr[:, k] * prior_units[k]

        samples['K'] = samples_arr[:, k + 1] * prior_units[-1]  # jitter unit

        for i in range(self.params.poly_trend):
            _unit = prior_units[-1] / u.day**i  # HACK: jitter unit per day
            samples['v'+str(i)] = samples_arr[:, k + 2 + i] * _unit

        if return_logprobs:
            return samples, ln_prior, ln_like

        else:
            return samples

    def _rejection_sample_from_cache(self, data, n_prior_samples, cache_file,
                                     start_idx, seed, return_logprobs=False):
        """Perform The Joker's rejection sampling on a cache file containing
        prior samples. This is meant to be used internally.
        """

        # Get indices of good samples from the cache file
        # TODO: I have some implementation questions about whether this should
        #   return a boolean array (in which case I need to process all
        #   likelihood values) or an array of integers...Right now,
        #   _marginal_ll_worker has to return the values because we then compare
        #   with the maximum value of the likelihood
        marg_lls = compute_likelihoods(n_prior_samples, cache_file, start_idx,
                                       data, self.params, pool=self.pool,
                                       n_batches=self.n_batches)
        good_samples_idx = get_good_sample_indices(marg_lls, seed=seed)

        if len(good_samples_idx) == 0:
            logger.error("Failed to find any good samples!")
            self.pool.close()
            sys.exit(1)

        n_good = len(good_samples_idx)
        s_or_not = 's' if n_good > 1 else ''
        logger.info("{0} good sample{1} after rejection sampling"
                    .format(n_good, s_or_not))

        # For samples that pass the rejection step, we now have their indices
        # in the prior cache file. Here, we read the actual values:
        result = sample_indices_to_full_samples(
            good_samples_idx, cache_file, data, self.params,
            pool=self.pool, global_seed=seed, return_logprobs=return_logprobs)

        return result

    def _validate_prior_cache(self, n_prior_samples, prior_cache_file):
        """Internal method used to either validate the prior cache file, or
        create one based on the number of samples input.
        """

        if n_prior_samples is None and prior_cache_file is None:
            raise ValueError("You either have to specify the number of prior "
                             "samples to generate, or a path to an HDF5 file "
                             "containing cached prior samples.")

        if prior_cache_file is not None:
            # read prior units from cache file
            with h5py.File(prior_cache_file, 'r') as f:
                if n_prior_samples is None:
                    n_prior_samples = len(f['samples'])

                # TODO: also validate keys in cache?

            cache_exists = True
            logger.log(1, "Prior cache file found and validated.")

        else:
            cache_exists = False
            logger.log(1, "Prior cache file not found or invalid.")

        return n_prior_samples, cache_exists

    def rejection_sample(self, data, n_prior_samples=None,
                         prior_cache_file=None, return_logprobs=False,
                         start_idx=0):
        """Run The Joker's rejection sampling on prior samples to get posterior
        samples for the input data.

        You must either specify the number of prior samples to generate and
        use for rejection sampling, ``n_prior_samples``, or the path to a file
        containing prior samples, ``prior_cache_file``.

        Parameters
        ----------
        data : `~thejoker.data.RVData`
            The radial velocity data.
        n_prior_samples : int (optional)
            If ``prior_cache_file`` is not specified, this sets the number of
            prior samples to generate and use to do the rejection sampling. If
            ``prior_cache_file`` is specified, this sets the number of prior
            samples to load from the cache file.
        prior_cache_file : str (optional)
            A path to an HDF5 cache file containing prior samples. TODO: more
            information
        return_logprobs : bool (optional)
            Also return the log-probabilities.
        start_idx : int (optional)
            Index to start reading from in the prior cache file.

        """

        # validate input data
        if not isinstance(data, RVData):
            raise TypeError("Input data must be an RVData instance, not '{0}'"
                            .format(type(data)))

        # compute full parameter vectors for all good samples
        if self._rnd_passed:
            seed = self.random_state.randint(np.random.randint(2**16))
        else:
            seed = None

        n_prior_samples, cache_exists = self._validate_prior_cache(
            n_prior_samples, prior_cache_file)

        if cache_exists:
            with h5py.File(prior_cache_file, 'r') as f:
                prior_units = [u.Unit(uu) for uu in f.attrs['units']]

            result = self._rejection_sample_from_cache(
                data, n_prior_samples, prior_cache_file, start_idx, seed=seed,
                return_logprobs=return_logprobs)

        else:
            with tempfile.NamedTemporaryFile(mode='r+',
                                             dir=self.tempfile_path) as f:
                prior_cache_file = f.name

                # first do prior sampling, cache to temporary file
                prior_samples, lnp = self.sample_prior(size=n_prior_samples,
                                                       return_logprobs=True)
                prior_units = save_prior_samples(prior_cache_file,
                                                 prior_samples,
                                                 data.rv.unit,
                                                 ln_prior_probs=lnp)

                result = self._rejection_sample_from_cache(
                    data, n_prior_samples, prior_cache_file, start_idx,
                    seed=seed, return_logprobs=return_logprobs)

        return self._unpack_full_samples(result, prior_units, t0=data.t0,
                                         return_logprobs=return_logprobs)

    def iterative_rejection_sample(self, data, n_requested_samples,
                                   prior_cache_file=None, n_prior_samples=None,
                                   return_logprobs=False, magic_fudge=128):
        """This is an experimental sampling method that adaptively generates
        posterior samples given a large library of prior samples. The advantage
        of this function over the standard ``rejection_sample`` method is that
        it will try to adaptively figure out how many prior samples it needs to
        evaluate the likelihood at in order to return the desired number of
        posterior samples.

        Parameters
        ----------
        data : `~thejoker.data.RVData`
            The radial velocity data.
        n_requested_samples : int (optional)
            The number of posterior samples desired.
        prior_cache_file : str
            A path to an HDF5 cache file containing prior samples.
        n_prior_samples : int (optional)
            The maximum number of prior samples to use.
        return_logprobs : bool (optional)
            Also return the log-probabilities.
        magic_fudge : int (optional)
            A magic fudge factor to use when adaptively determining the number
            of prior samples to evaluate at. Larger numbers make the trial
            batches larger more rapidly.
        """

        # validate input data
        if not isinstance(data, RVData):
            raise TypeError("Input data must be an RVData instance, not '{}'"
                            .format(type(data)))

        # a bit of a hack to make the K, v0 samples deterministic
        if self._rnd_passed:
            seed = self.random_state.randint(np.random.randint(2**16))
        else:
            seed = None

        # HACK: must start from 0
        start_idx = 0

        n_prior_samples, cache_exists = self._validate_prior_cache(
            n_prior_samples, prior_cache_file)

        # There are some magic numbers below used to control how fast the
        # iterative batches grow in size
        safety_factor = 2  # MAGIC NUMBER
        n_process = magic_fudge * n_requested_samples  # MAGIC NUMBER

        if n_process > n_prior_samples:
            raise ValueError("Prior sample library not big enough! For "
                             "iterative sampling, you have to have at least "
                             "magic_fudge * n_requested_samples samples in the "
                             "prior samples cache file. You have {0}"
                             .format(n_prior_samples))

        all_marg_lls = np.array([])

        # TODO: it's a little...unclean to always make a tempfile

        if cache_exists:
            logger.log(1, "Cache file exists at: {0}"
                       .format(prior_cache_file))
            with h5py.File(prior_cache_file, 'r') as f:
                prior_units = [u.Unit(uu) for uu in f.attrs['units']]
            close_f = False

        else:
            f = tempfile.NamedTemporaryFile(mode='r+', dir=self.tempfile_path)
            close_f = True
            prior_cache_file = f.name
            logger.log(1, "Cache file not found - creating prior samples "
                       "and saving them to: {0}".format(prior_cache_file))

            # first do prior sampling, cache to temporary file
            prior_samples, lnp = self.sample_prior(size=n_prior_samples,
                                                   return_logprobs=True)
            prior_units = save_prior_samples(f.name, prior_samples,
                                             data.rv.unit,
                                             ln_prior_probs=lnp)

        maxiter = 128
        for i in range(maxiter):  # we just need to iterate for a long time
            logger.log(1, "The Joker iteration {0}, computing {1} "
                       "likelihoods".format(i, n_process))
            marg_lls = compute_likelihoods(n_process, prior_cache_file,
                                           start_idx, data, self.params,
                                           pool=self.pool,
                                           n_batches=self.n_batches)

            all_marg_lls = np.concatenate((all_marg_lls, marg_lls))

            good_samples_idx = get_good_sample_indices(all_marg_lls,
                                                       seed=seed)

            if len(good_samples_idx) == 0:
                # self.pool.close()
                raise RuntimeError("Failed to find any good samples!")

            n_good = len(good_samples_idx)
            logger.log(1, "{0} good samples after rejection sampling"
                       .format(n_good))

            if len(good_samples_idx) >= n_requested_samples:
                logger.debug("Enough samples found! {0}"
                             .format(len(good_samples_idx)))
                break

            start_idx += n_process

            n_ll_evals = len(all_marg_lls)
            n_need = n_requested_samples - n_good
            n_process = int(safety_factor * n_need / n_good * n_ll_evals)

            if start_idx + n_process > n_prior_samples:
                n_process = n_prior_samples - start_idx

            if n_process <= 0:
                break

        else:
            # We should never get here!!
            raise RuntimeError("Hit maximum number of iterations!")

        result = sample_indices_to_full_samples(
            good_samples_idx[:n_requested_samples], prior_cache_file, data,
            self.params, pool=self.pool, global_seed=seed,
            return_logprobs=return_logprobs)

        if close_f:
            logger.log(1, "Closing prior cache tempfile")
            f.close()

        return self._unpack_full_samples(result, prior_units, t0=data.t0,
                                         return_logprobs=return_logprobs)

    # ========================================================================
    # MCMC

    def mcmc_sample(self, data, samples0, n_steps=1024,
                    n_walkers=256, n_burn=8192, return_sampler=False,
                    ball_scale=1E-5):
        """Run standard MCMC (using `emcee <http://emcee.readthedocs.io/>`_) to
        generate posterior samples in orbital parameters.

        Parameters
        ----------
        data : `~thejoker.RVData`
            The data to fit orbits to.
        samples0 : `~thejoker.JokerSamples`
            This can either be (a) a single sample to use as initial conditions
            for the MCMC walkers, or (b) a set of samples, in which case the
            mean of the samples will be used as initial conditions.
        n_steps : int
            The number of MCMC steps to run for.
        n_walkers : int (optional)
            The number of walkers to use in the ``emcee`` ensemble.
        n_burn : int (optional)
            If specified, the number of steps to burn in for.
        return_sampler : bool (optional)
            Also return the sampler object.

        Returns
        -------
        model : `~thejoker.TheJokerMCMCModel`
        samples : `~thejoker.JokerSamples`
            The posterior samples.
        sampler : `emcee.EnsembleSampler`
            If ``return_sampler == True``.
        """
        import emcee

        if not isinstance(samples0, JokerSamples):
            raise TypeError('Input samples initial position must be ')

        model = TheJokerMCMCModel(joker_params=self.params, data=data)

        if len(samples0) > 1:
            # samples0 = samples0.median()
            samples0 = samples0[0] # TODO:
        t0 = samples0.t0

        samples0 = model._strip_units(samples0)
        all_samples0 = dict()
        for i, k in enumerate(samples0.keys()):
            all_samples0[k] = np.random.normal(samples0[k], ball_scale,
                                               size=n_walkers)

            if k in ['e', 'K', 'jitter']:
                all_samples0[k] = np.abs(all_samples0[k])

        p0 = np.squeeze(model.pack_samples(all_samples0, strip_units=False)).T
        n_dim = p0.shape[1]

        sampler = emcee.EnsembleSampler(n_walkers, n_dim, model,
                                        pool=self.pool)

        if n_burn is not None and n_burn > 0:
            logger.debug('Burning in MCMC for {0} steps...'.format(n_burn))
            time0 = time.time()
            pos, *_ = sampler.run_mcmc(p0, n_burn)
            logger.debug('...time spent burn-in: {0}'.format(time.time()-time0))

            p0 = pos
            sampler.reset()

        logger.debug('Running MCMC for {0} steps...'.format(n_steps))
        time0 = time.time()
        _ = sampler.run_mcmc(p0, n_steps)
        logger.debug('...time spent sampling: {0}'.format(time.time()-time0))

        acc_frac = sampler.acceptance_fraction
        if np.percentile(acc_frac, 10) < 0.1:
            logger.warning('Walkers have low acceptance fractions: 10/50/90 '
                           'percentiles = {0:.2f}, {1:.2f}, {2:.2f}'
                           .format(*np.percentile(acc_frac, [10, 50, 90])))

        samples = model.unpack_samples(sampler.chain[:, -1])
        samples.t0 = t0

        if return_sampler:
            return model, samples, sampler

        else:
            return model, samples



# Third-party
import astropy.units as u
import h5py
import numpy as np
import schwimmbad

# Package
from ..multiproc_helpers import (get_good_sample_indices, compute_likelihoods,
                                 sample_indices_to_full_samples, chunk_tasks)
from .helpers import FakeData


def test_chunk_tasks():
    N = 10000
    start_idx = 1103
    tasks = chunk_tasks(N, n_batches=16, start_idx=start_idx)
    assert tasks[0][0][0] == start_idx
    assert tasks[-1][0][1] == N+start_idx

    # try with an array:
    tasks = chunk_tasks(N, n_batches=16, start_idx=start_idx,
                        arr=np.random.random(size=8*N))
    n_tasks = sum([tasks[i][0].size for i in range(len(tasks))])
    assert n_tasks == N


class TestMultiproc(object):

    # TODO: this is bad to copy pasta from test_likelihood.py
    def truths_to_nlp(self, truths):
        # P, M0, ecc, omega
        P = truths['P'].to(u.day).value
        M0 = truths['M0'].to(u.radian).value
        ecc = truths['e']
        omega = truths['omega'].to(u.radian).value
        return np.array([P, M0, ecc, omega, 0.])

    def setup(self):
        d = FakeData()
        self.data = d.datasets
        self.joker_params = d.params
        self.truths = d.truths

    def test_multiproc_helpers(self, tmpdir):
        prior_samples_file = str(tmpdir.join('prior-samples.h5'))
        pool = schwimmbad.SerialPool()

        data = self.data['circ_binary']
        joker_params = self.joker_params['circ_binary']
        truth = self.truths['circ_binary']
        nlp = self.truths_to_nlp(truth)

        # write some nonsense out to the prior file
        n = 8192
        P = np.random.uniform(nlp[0]-2., nlp[0]+2., n)
        M0 = np.random.uniform(0, 2*np.pi, n)
        ecc = np.zeros(n)
        omega = np.zeros(n)
        jitter = np.zeros(n)
        samples = np.vstack((P,M0,ecc,omega,jitter)).T

        # TODO: use save_prior_samples here

        with h5py.File(prior_samples_file) as f:
            f['samples'] = samples

        lls = compute_likelihoods(n, prior_samples_file, 0, data,
                                  joker_params, pool)
        idx = get_good_sample_indices(lls)
        assert len(idx) >= 1

        lls = compute_likelihoods(n, prior_samples_file, 0, data,
                                  joker_params, pool, n_batches=13)
        idx = get_good_sample_indices(lls)
        assert len(idx) >= 1

        full_samples = sample_indices_to_full_samples(idx, prior_samples_file,
                                                      data, joker_params, pool)
        print(full_samples)



# Third-party
from astropy.utils.misc import isiterable
import astropy.units as u
import numpy as np

__all__ = ['JokerParams']


class JokerParams:
    """

    Parameters
    ----------
    P_min : `astropy.units.Quantity` [time]
        Lower bound on prior over period, the smallest period considered.
    P_max : `astropy.units.Quantity` [time]
        Upper bound on prior over period, the largest period considered.
    jitter : `~astropy.units.Quantity` [speed], tuple (optional)
        Represents additional Gaussian noise in the RV signal. Default
        is to fix the value of the jitter to 0. To fix the jitter to a
        different value, pass in a single `~astropy.units.Quantity`
        object. The Joker also supports inferring the jitter as an
        additional non-linear parameter. Currently, the only prior
        pdf supported for doing this is a Gaussian in natural-log of
        the jitter squared--that is,
        :math:`p(a) = \mathcal{N}(a|\mu,\sigma)` where
        :math:`a = \log s^2`. The (dimensionless) mean and standard
        deviation of this prior can also be passed in to this argument
        by passing a length-2 tuple of numbers. If you do this, you must
        also pass in a unit for the jitter using the ``jitter_unit`` arg.
    jitter_unit : `~astropy.units.UnitBase`
        If sampling over the jitter as an extra non-linear parameter,
        you must also specify the units of the jitter prior. See note
        above about the ``jitter`` argument.
    poly_trend : int, optional
        If specified, sample over a polynomial velocity trend with the specified
        number of coefficients. For example, ``poly_trend=3`` will sample over
        parameters of a long-term quadratic velocity trend. Default is 1, just a
        constant velocity shift.
    linear_par_Vinv : array_like, optional
        Inverse variance matrix that specifies the Gaussian prior on the linear
        parameters, i.e., the semi-amplitude and velocity trend paramters. The
        units must be in inverse, squared ``v_unit`` and ``v_unit/day^n`` where
        ``v_unit`` is the jitter velocity unit, and ``day^n`` corresponds to
        each polynomial trend coefficient.
    anomaly_tol : float (optional)
        Convergence tolerance passed to
        :func:`twobody.eccentric_anomaly_from_mean_anomaly`.
        Arbitrarily set to 1E-10 by default.
    anomaly_maxiter : float (optional)
        Maximum number of iterations passed to
        :func:`twobody.eccentric_anomaly_from_mean_anomaly`.
        Arbitrarily set to 128 by default.

    Examples
    --------

        >>> import astropy.units as u
        >>> pars = JokerParams(P_min=8*u.day, P_max=8192*u.day,
        ...                    jitter=5.*u.m/u.s) # fix jitter to 5 m/s
        >>> pars = JokerParams(P_min=8*u.day, P_max=8192*u.day,
        ...                    jitter=(1., 2.), jitter_unit=u.m/u.s) # specify jitter prior

    """
    @u.quantity_input(P_min=u.day, P_max=u.day)
    def __init__(self, P_min, P_max,
                 jitter=None, jitter_unit=None,
                 poly_trend=1, linear_par_Vinv=None,
                 anomaly_tol=1E-10, anomaly_maxiter=128):

        # the names of the default parameters
        self.default_params = ['P', 'M0', 'e', 'omega', 'jitter', 'K']

        self.poly_trend = int(poly_trend)
        self.default_params += ['v{0}'.format(i)
                                for i in range(self.poly_trend)]

        self.P_min = P_min
        self.P_max = P_max
        self.anomaly_tol = float(anomaly_tol)
        self.anomaly_maxiter = int(anomaly_maxiter)

        # K, then the linear parameters
        _n_linear = 1 + self.poly_trend
        if linear_par_Vinv is None:
            self.linear_par_Vinv = 1e-10 * np.eye(_n_linear)
        else:
            self.linear_par_Vinv = np.array(linear_par_Vinv)
            if self.linear_par_Vinv.shape != (_n_linear, _n_linear):
                raise ValueError("Linear parameter inverse variance prior must "
                                 "have shape ({0}, {0})".format(_n_linear))

        # validate the input jitter specification
        if jitter is None:
            jitter = 0 * u.km/u.s

        if isiterable(jitter):
            if len(jitter) != 2:
                raise ValueError("If specifying parameters for the jitter prior, you "
                                 "must pass in a length-2 container containing the "
                                 "mean and standard deviation of the Gaussian over "
                                 "log(jitter^2)")

            if jitter_unit is None or not isinstance(jitter_unit, u.UnitBase):
                raise TypeError("If specifying parameters for the jitter prior, you "
                                "must also specify the units of the jitter for "
                                "evaluating the prior as an astropy.units.UnitBase "
                                "instance.")

            self._fixed_jitter = False
            self._jitter_unit = jitter_unit
            self.jitter = jitter

        else:
            self._fixed_jitter = True
            self._jitter_unit = jitter.unit
            self.jitter = jitter

    @property
    def num_params(self):
        n = len(self.default_params)
        return n

    # --- All old shit below here that needs to be moved ---

    # @classmethod
    # def get_labels(cls, units=None):
    #     _u = dict()
    #     if units is None:
    #         _u = cls._name_to_unit

    #     else:
    #         for k,unit in cls._name_to_unit.items():
    #             if k in units:
    #                 _u[k] = units[k]
    #             else:
    #                 _u[k] = unit

    #     _labels = [
    #         r'$\ln (P/1\,${}$)$'.format(_u['P'].long_names[0]),
    #         '$e$',
    #         r'$\omega$ [{}]'.format(_u['omega']),
    #         r'$\phi_0$ [{}]'.format(_u['omega']),
    #         r'$\ln (s/1\,${}$)$'.format(_u['jitter'].to_string(format='latex_inline')),
    #         r'$K$ [{}]'.format(_u['K'].to_string(format='latex_inline')),
    #         '$v_0$ [{}]'.format(_u['v0'].to_string(format='latex_inline'))
    #     ]

    #     return _labels

    # @classmethod
    # def from_hdf5(cls, f):
    #     kwargs = dict()
    #     if isinstance(f, six.string_types):
    #         with h5py.File(f, 'r') as g:
    #             for key in cls._name_to_unit.keys():
    #                 kwargs[key] = quantity_from_hdf5(g, key)

    #     else:
    #         for key in cls._name_to_unit.keys():
    #             kwargs[key] = quantity_from_hdf5(f, key)

    #     return cls(**kwargs)

    # def to_hdf5(self, f):
    #     if isinstance(f, six.string_types):
    #         with h5py.File(f, 'a') as g:
    #             for key in self._name_to_unit.keys():
    #                 quantity_to_hdf5(g, key, getattr(self, key))

    #     else:
    #         for key in self._name_to_unit.keys():
    #             quantity_to_hdf5(f, key, getattr(self, key))

    # def pack(self, units=None, plot_transform=False):
    #     """
    #     Pack the orbital parameters into a single array structure
    #     without associated units. The components will have units taken
    #     from the unit system defined in `thejoker.units.usys`.

    #     Parameters
    #     ----------
    #     units : dict (optional)
    #     plot_transform : bool (optional)

    #     Returns
    #     -------
    #     pars : `numpy.ndarray`
    #         A single 2D array containing the parameter values with no
    #         units. Will have shape ``(n,6)``.

    #     """
    #     if units is None:
    #         all_samples = np.vstack([getattr(self, "_{}".format(key))
    #                                  for key in self._name_to_unit.keys()]).T

    #     else:
    #         all_samples = np.vstack([getattr(self, format(key)).to(units[key]).value
    #                                  for key in self._name_to_unit.keys()]).T

    #     if plot_transform:
    #         # ln P in plots:
    #         idx = list(self._name_to_unit.keys()).index('P')
    #         all_samples[:,idx] = np.log(all_samples[:,idx])

    #         # ln s in plots:
    #         idx = list(self._name_to_unit.keys()).index('jitter')
    #         all_samples[:,idx] = np.log(all_samples[:,idx])

    #     return all_samples

    # @classmethod
    # def unpack(cls, pars):
    #     """
    #     Unpack a 2D array structure containing the orbital parameters
    #     without associated units. Should have shape ``(n,6)`` where ``n``
    #     is the number of parameters.

    #     Returns
    #     -------
    #     p : `~thejoker.celestialmechanics.OrbitalParams`

    #     """
    #     kw = dict()
    #     par_arr = np.atleast_2d(pars).T
    #     for i,key in enumerate(cls._name_to_unit.keys()):
    #         kw[key] = par_arr[i] * cls._name_to_unit[key]

    #     return cls(**kw)

    # def copy(self):
    #     return self.__copy__()

    # def rv_orbit(self, index=None):
    #     """
    #     Get a `~thejoker.celestialmechanics.SimulatedRVOrbit` instance
    #     for the orbital parameters with index ``i``.

    #     Parameters
    #     ----------
    #     index : int (optional)

    #     Returns
    #     -------
    #     orbit : `~thejoker.celestialmechanics.SimulatedRVOrbit`
    #     """
    #     from .celestialmechanics_class import SimulatedRVOrbit

    #     if index is None and len(self._P) == 1: # OK
    #         index = 0

    #     elif index is None and len(self._P) > 1:
    #         raise IndexError("You must specify the index of the set of paramters to get an "
    #                          "orbit for!")

    #     i = index
    #     return SimulatedRVOrbit(self[i])



# Third-party
from astropy.time import Time
import astropy.units as u
import numpy as np

# Project
from .log import log as logger

__all__ = ['RVData']


class RVData:
    """
    Time-domain radial velocity measurements for a single target.

    Parameters
    ----------
    t : `~astropy.time.Time`, array_like
        Array of measurement times. Either as an array of BMJD values
        or as an Astropy time object.
    rv : `~astropy.units.Quantity` [speed]
        Radial velocity (RV) measurements.
    stddev : `~astropy.units.Quantity` [speed] (optional)
        Standard deviation for each RV measurement. Specify this or ``ivar``.
    ivar : `~astropy.units.Quantity` [1/speed^2] (optional)
        Inverse variance for each RV measurement. Specify this or ``stddev``.
    metadata : any (optional)
        Any metadata associated with the object.
    t0 : numeric (optional) [day]
        A reference time. Default is to use the minimum time in barycentric MJD
        (days).

    """
    @u.quantity_input(rv=u.km/u.s)
    def __init__(self, t, rv, ivar=None, stddev=None,
                 metadata=None, t0=None):

        # For speed, many of the attributes are saved without units and only
        #   returned with units if asked for.
        if isinstance(t, Time):
            _t_bmjd = t.tcb.mjd
        else:
            _t_bmjd = np.atleast_1d(t)
        self._t_bmjd = _t_bmjd

        self.rv = np.atleast_1d(rv)

        # parse input specification of errors
        self._has_err = True
        if ivar is None and stddev is None:
            self.ivar = np.full_like(self.rv.value, np.nan) * self.rv.unit
            self._has_err = False

        elif ivar is not None and stddev is not None:
            raise ValueError("You must pass in 'ivar' or 'stddev', not both.")

        elif ivar is not None:
            if not hasattr(ivar, 'unit'):
                raise TypeError("ivar must be an Astropy Quantity object")
            elif not ivar.unit.is_equivalent(1/self.rv.unit**2):
                raise u.UnitsError("ivar must have same unit type as RV^-2")

            self.ivar = ivar.to(1/self.rv.unit**2)

        elif stddev is not None:
            if not hasattr(stddev, 'unit'):
                raise TypeError("stddev must be an Astropy Quantity object!")
            elif not stddev.unit.is_equivalent(self.rv.unit):
                raise u.UnitsError("stddev must have same unit type as RV")

            self.ivar = 1 / stddev.to(self.rv.unit)**2
        self.ivar = np.atleast_1d(self.ivar)

        # make sure shapes are consistent
        if self._t_bmjd.shape != self.rv.shape or self.rv.shape != self.ivar.shape:
            raise ValueError("Shape of input time, RV, and errors must be consistent! "
                             "({} vs {} vs {})".format(self._t_bmjd.shape,
                                                       self.rv.shape,
                                                       self.ivar.shape))
        # filter out NAN or INF data points
        idx = np.isfinite(self._t_bmjd) & np.isfinite(self.rv)

        if self._has_err:
            idx &= np.isfinite(self.ivar) & (self.ivar.value > 0)

        if idx.sum() < len(self.rv):
            logger.info("Filtering {} NaN/Inf data points".format(len(self.rv) - idx.sum()))

        self._t_bmjd = self._t_bmjd[idx]
        self.rv = self.rv[idx]
        self.ivar = self.ivar[idx]

        # sort on times
        idx = self._t_bmjd.argsort()
        self._t_bmjd = self._t_bmjd[idx]
        self.rv = self.rv[idx]
        self.ivar = self.ivar[idx]

        # metadata can be anything
        self.metadata = metadata

        # if no offset is provided, subtract the minimum time
        if t0 is None:
            t0 = self.t.min()

        if not isinstance(t0, Time):
            raise TypeError('If a reference time t0 is specified, it must '
                            'be an astropy.time.Time object.')

        self.t0 = t0
        self._t0_bmjd = self.t0.tcb.mjd

    @property
    def t(self):
        """
        The times of each observation.

        Returns
        -------
        t : `~astropy.time.Time`
            An Astropy Time object for all times.
        """
        return Time(self._t_bmjd, scale='tcb', format='mjd')

    def phase(self, P, t0=None):
        """
        Convert time to a phase.

        By default, the phase is relative to the internal reference epoch,
        ``t0``, but a new epoch can also be specified to this method.

        Parameters
        ----------
        P : `~astropy.units.Quantity` [time]
            The period.
        t0 : `~astropy.time.Time` (optional)
            Default uses the internal reference epoch. Use this to compute the
            phase relative to some other epoch

        Returns
        -------
        phase : `~numpy.ndarray`
            The dimensionless phase of each observation.

        """
        if t0 is None:
            t0 = self.t0
        return ((self.t - t0) / P) % 1.

    @property
    def stddev(self):
        """
        The Gaussian error for each radial velocity measurement.

        Returns
        -------
        stddev : `~astropy.units.Quantity` [speed]
            An Astropy Quantity with velocity (speed) units.
        """
        return 1 / np.sqrt(self.ivar)

    # ---

    def plot(self, ax=None, rv_unit=None, time_format='mjd', phase_fold=None,
             relative_to_t0=False, add_labels=True, **kwargs):
        """
        Plot the data points.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes` (optional)
            The matplotlib axes object to draw on (default is to grab
            the current axes object using `~matplotlib.pyplot.gca`).
        rv_unit : `~astropy.units.UnitBase` (optional)
            Display the radial velocities with a different unit
            (default uses whatever unit was passed on creation).
        time_format : str, callable (optional)
            The time format to use for the x-axis. This can either be
            a string, in which case it is assumed to be an attribute of
            the `~astropy.time.Time` object, or it can be a callable (e.g.,
            function) that does more complex things (for example:
            ``time_format=lambda t: t.datetime.day``).
        phase_fold : bool (optional)
            Plot the phase instead of the time by folding on the specified
            period.
        relative_to_t0 : bool (optional)
            Plot the time relative to the reference epoch, ``t0``.
        add_labels : bool (optional)
            Add labels to the figure.
        **kwargs
            All other keyword arguments are passed to the
            `~matplotlib.pyplot.errorbar` (if errors were provided) or
            `~matplotlib.pyplot.plot` (if no errors provided) call.

        """
        if ax is None:
            import matplotlib.pyplot as plt
            ax = plt.gca()

        if rv_unit is None:
            rv_unit = self.rv.unit

        # some default stylings
        style = kwargs.copy()
        style.setdefault('linestyle', 'none')
        style.setdefault('alpha', 1.)
        style.setdefault('marker', 'o')
        style.setdefault('color', 'k')
        style.setdefault('ecolor', '#666666')
        style.setdefault('elinewidth', 1)

        if callable(time_format):
            t = time_format(self.t)
            t0 = time_format(self.t0)
        else:
            t = getattr(self.t, time_format)
            t0 = getattr(self.t0, time_format)

        if relative_to_t0:
            t = t - t0

        if phase_fold:
            t = (t / phase_fold.to(u.day).value) % 1

        if self._has_err:
            ax.errorbar(t, self.rv.to(rv_unit).value,
                        self.stddev.to(rv_unit).value,
                        **style)
        else:
            style.pop('ecolor')
            style.pop('elinewidth')
            ax.plot(t, self.rv.to(rv_unit).value, **style)

        if add_labels:
            ax.set_xlabel('time [BMJD]')
            ax.set_ylabel('RV [{:latex_inline}]'.format(rv_unit))

        return ax

    # copy methods
    def __copy__(self):
        return self.__class__(t=self.t.copy(),
                              rv=self.rv.copy(),
                              ivar=self.ivar.copy())

    def copy(self):
        return self.__copy__()

    def __getitem__(self, slc):
        return self.__class__(t=self.t.copy()[slc],
                              rv=self.rv.copy()[slc],
                              ivar=self.ivar.copy()[slc])

    def __len__(self):
        return len(self.rv.value)

    def to_hdf5(self, file_or_path):
        """
        Write data to an HDF5 file.

        Parameters
        ----------
        file_or_path : str, `h5py.File`, `h5py.Group`
        """

        import h5py
        if isinstance(file_or_path, str):
            f = h5py.File(file_or_path, 'w')
            close = True

        else:
            f = file_or_path
            close = False

        d = f.create_dataset('mjd', data=self.t.tcb.mjd)
        d.attrs['format'] = 'mjd'
        d.attrs['scale'] = 'tcb'

        d = f.create_dataset('rv', data=self.rv.value)
        d.attrs['unit'] = str(self.rv.unit)

        d = f.create_dataset('rv_err', data=self.stddev.value)
        d.attrs['unit'] = str(self.stddev.unit)

        if close:
            f.close()

    @classmethod
    def from_hdf5(cls, file_or_path):
        """
        Read data to an HDF5 file.

        Parameters
        ----------
        file_or_path : str, `h5py.File`, `h5py.Group`
        """

        import h5py
        if isinstance(file_or_path, str):
            f = h5py.File(file_or_path, 'r')
            close = True

        else:
            f = file_or_path
            close = False

        t = f['mjd']
        rv = f['rv'][:] * u.Unit(f['rv'].attrs['unit'])
        stddev = f['rv_err'][:] * u.Unit(f['rv_err'].attrs['unit'])

        if close:
            f.close()

        return cls(t=t, rv=rv, stddev=stddev)



# Third-party
import h5py
import numpy as np

# Project
from ..log import log
from .likelihood import (get_ivar, design_matrix, tensor_vector_scalar,
                         marginal_ln_likelihood)
from .fast_likelihood import (batch_marginal_ln_likelihood,
                              batch_get_posterior_samples)

__all__ = ['compute_likelihoods', 'get_good_sample_indices',
           'sample_indices_to_full_samples']


def chunk_tasks(n_tasks, n_batches, arr=None, args=None, start_idx=0):
    """Split the tasks into some number of batches to sent out to MPI workers.

    Parameters
    ----------
    n_tasks : int
        The total number of tasks to divide.
    n_batches : int
        The number of batches to split the tasks into. Often, you may want to do
        ``n_batches=pool.size`` for equal sharing amongst MPI workers.
    arr : iterable (optional)
        Instead of returning indices that specify the batches, you can also
        directly split an array into batches.
    args : iterable (optional)
        Other arguments to add to each task.
    start_idx : int (optional)
        What index in the tasks to start from?

    """
    if args is None:
        args = []
    args = list(args)

    tasks = []
    if n_batches > 0 and n_tasks > n_batches:
        # chunk by the number of batches, often the pool size
        base_chunk_size = n_tasks // n_batches
        rmdr = n_tasks % n_batches

        i1 = start_idx
        for i in range(n_batches):
            i2 = i1 + base_chunk_size
            if i < rmdr:
                i2 += 1

            if arr is None: # store indices
                tasks.append([(i1, i2), i1] + args)

            else: # store sliced array
                tasks.append([arr[i1:i2], i1] + args)

            i1 = i2

    else:
        if arr is None: # store indices
            tasks.append([(start_idx, n_tasks+start_idx), start_idx] + args)

        else: # store sliced array
            tasks.append([arr[start_idx:n_tasks+start_idx], start_idx] + args)

    return tasks


def _marginal_ll_worker(task):
    """
    Compute the marginal log-likelihood, i.e. the likelihood integrated over
    the linear parameters. This is meant to be ``map``ped using a processing
    pool` within the functions below and is not supposed to be in the
    public API.

    Parameters
    ----------
    task : iterable
        An array containing the indices of samples to be operated on, the
        filename containing the prior samples, and the data.

    Returns
    -------
    ll : `numpy.ndarray`
        Array of log-likelihood values.

    """
    start_stop, chunk_index, prior_cache_file, data, jparams = task

    # read a chunk of the prior samples
    with h5py.File(prior_cache_file, 'r') as f:
        chunk = np.array(f['samples'][start_stop[0]:start_stop[1]])

    chunk = chunk.astype(np.float64)

    # memoryview is returned
    ll = batch_marginal_ln_likelihood(chunk, data, jparams)
    return np.array(ll)


def compute_likelihoods(n_prior_samples, prior_cache_file, start_idx, data,
                        joker_params, pool, n_batches=None):
    """
    Return the indices of 'good' samples by computing the log-likelihood
    for ``n_prior_samples`` prior samples and doing rejection sampling.

    For speed when parallelizing, this accepts a filename for an HDF5
    that contains the prior samples, splits up the samples based on the
    number of processes / MPI workers, and only distributes the indices
    for each worker to read. This limits the amount of data that needs
    to be passed around.

    Parameters
    ----------
    n_prior_samples : int
        The number of prior samples to use.
    prior_cache_file : str
        Path to an HDF5 file containing the prior samples.
    start_idx : int
        Index to start reading prior samples from in the prior cache file.
    data : `~thejoker.data.RVData`
        An instance of ``RVData`` with the data we're modeling.
    joker_params : `~thejoker.sampler.params.JokerParams`
        A specification of the parameters to use.
    pool : `~schwimmbad.pool.BasePool` or subclass
        An instance of a processing pool - must have a ``.map()`` method.
    n_batches : int (optional)
        How many batches to divide the work into. Defaults to ``pool.size``.

    Returns
    -------
    samples_idx : `numpy.ndarray`
        An array of integers for the prior samples that pass
        rejection sampling.

    TODO
    ----
    - The structure of this function is ok for most cluster-like machines
        up to ~2**28 samples or so. Then it becomes an issue that I keep
        the likelihood values in memory. We may want to be able to cache
        the likelihood values instead?

    """
    args = [prior_cache_file, data, joker_params]
    if n_batches is None:
        n_batches = pool.size
    tasks = chunk_tasks(n_prior_samples, n_batches=n_batches, args=args,
                        start_idx=start_idx)

    results = [r for r in pool.map(_marginal_ll_worker, tasks)]
    marg_ll = np.concatenate(results)

    if len(marg_ll) != n_prior_samples:
        raise RuntimeError("Unexpected failure: number of likelihoods "
                           "returned from workers does not match number sent "
                           "out to workers.")

    return marg_ll


def get_good_sample_indices(marg_ll, seed=None):
    """Return the indices of 'good' samples from pre-computed values of the
    log-likelihood.

    Parameters
    ----------
    marg_ll : array_like
        Array of marginal log-likelihood values.
    seed : int (optional)
        Random number seed for uniform samples to use in rejection sampling.

    Returns
    -------
    samples_idx : `numpy.ndarray`
        An array of integers for the prior samples that pass
        rejection sampling.

    """

    # rejection sample using the marginal likelihood
    rnd = np.random.RandomState(seed)
    uu = rnd.uniform(size=len(marg_ll))
    good_samples_bool = uu < np.exp(marg_ll - marg_ll.max())
    good_samples_idx, = np.where(good_samples_bool)

    return good_samples_idx


def _sample_vector_worker(task):
    """
    This is meant to be
        ``map``ped using one of the ``Pool`` classes by the functions below and
        is not supposed to be in the public API.
    """

    (idx, chunk_index, prior_cache_file, data, joker_params, global_seed,
     return_logprobs) = task

    if global_seed is not None:
        seed = global_seed + chunk_index
        rnd = np.random.RandomState(seed)
        log.debug("worker with chunk {} has seed {}".format(idx[0], seed))

    else:
        rnd = np.random.RandomState()
        log.debug("worker with chunk {} not seeded".format(idx[0]))

    # read a chunk of the prior samples
    with h5py.File(prior_cache_file, 'r') as f:
        tmp = np.zeros(len(f['samples']), dtype=bool)
        tmp[idx] = True
        chunk = np.array(f['samples'][tmp, :])

        if return_logprobs:
            ln_prior = np.array(f['ln_prior_probs'][tmp])

    chunk = chunk.astype(np.float64)

    pars = batch_get_posterior_samples(chunk, data, joker_params, rnd,
                                       return_logprobs)
    if return_logprobs:
        pars = np.hstack((pars[:, :-1], ln_prior[:, None], pars[:, -1:]))
    return pars


def sample_indices_to_full_samples(good_samples_idx, prior_cache_file, data,
                                   joker_params, pool, global_seed=None,
                                   return_logprobs=False, n_batches=None):
    """
    Generate the full set of parameter values (linear + non-linear) for
    the nonlinear parameter prior samples that pass the rejection sampling.

    For speed when parallelizing, this accepts a filename for an HDF5
    that contains the prior samples, splits up the samples based on the
    number of processes / MPI workers, and only distributes the indices
    for each worker to read.

    Parameters
    ----------
    good_samples_idx : array_like
        The array of indices for the 'good' samples in the prior
        samples cache file.
    prior_cache_file : str
        Path to an HDF5 file containing the prior samples.
    data : `thejoker.data.RVData`
        An instance of ``RVData`` with the data we're modeling.
    joker_params : `~thejoker.sampler.params.JokerParams`
        A specification of the parameters to use.
    pool : `~schwimmbad.pool.BasePool` or subclass
        An instance of a processing pool - must have a ``.map()`` method.
    global_seed : int (optional)
        The global level random number seed.
    return_logprobs : bool (optional)
        Also return the log-probabilities of the prior samples.
    n_batches : int (optional)
        How many batches to divide the work into. Defaults to ``pool.size``.

    """

    n_samples = len(good_samples_idx)
    args = [prior_cache_file, data, joker_params, global_seed, return_logprobs]
    if n_batches is None:
        n_batches = pool.size
    tasks = chunk_tasks(n_samples, n_batches=n_batches, arr=good_samples_idx,
                        args=args)

    samples = [r for r in pool.map(_sample_vector_worker, tasks)]
    samples = np.concatenate(samples)

    assert len(samples) == n_samples
    samples = samples.reshape(-1, samples.shape[-1])

    if return_logprobs:
        # samples, ln(prior), ln(likelihood)
        return samples[:, :-2], samples[:, -2], samples[:, -1]

    else:
        return samples



from __future__ import absolute_import
from distutils.core import Extension
from astropy_helpers import setup_helpers

def get_extensions():
    exts = []

    from os import path
    try:
        # Get twobody path, or ignore failure -- for egg_info build
        import twobody
        twobody_path = path.dirname(twobody.__file__)
    except ImportError:
        twobody_path = None

    cfg = setup_helpers.DistutilsExtensionArgs()
    cfg['include_dirs'].append('numpy')

    if twobody_path is not None:
        cfg['include_dirs'].append(twobody_path)
        cfg['sources'].append(path.join(twobody_path,
                                        'src/twobody.c'))

    cfg['extra_compile_args'].append('--std=gnu99')
    cfg['sources'].append('thejoker/sampler/fast_likelihood.pyx')
    exts.append(Extension('thejoker.sampler.fast_likelihood', **cfg))

    return exts



# Third-party
import astropy.time as atime
import astropy.units as u
import numpy as np
import pytest

try:
    import matplotlib.pyplot as plt
    HAS_MPL = True
except:
    HAS_MPL = False

# Package
from ..sampler import JokerSamples
from ..plot import plot_rv_curves


@pytest.mark.skipif(not HAS_MPL, reason='matplotlib not installed')
def test_plot_rv_curves():

    pars = dict(P=[30.]*u.day, K=[10.]*u.km/u.s, e=[0.11239],
                omega=[0.]*u.radian, M0=[np.pi/2]*u.radian,
                v0=[150]*u.km/u.s)

    samples = JokerSamples()
    for k in pars:
        samples[k] = pars[k]

    t_grid = np.random.uniform(56000, 56500, 1024)
    t_grid.sort()

    fig, ax = plt.subplots(1, 1, figsize=(12,5))
    plot_rv_curves(samples, t_grid, ax=ax)



# Standard library
import warnings

# Third-party
from astropy.time import Time
import astropy.units as u
import numpy as np

__all__ = ['plot_rv_curves']


def plot_rv_curves(samples, t_grid=None, rv_unit=None, data=None,
                   ax=None, plot_kwargs=dict(), data_plot_kwargs=dict(),
                   add_labels=True, relative_to_t0=False):
    """
    Plot radial velocity curves for the input set of orbital parameter
    samples over the input grid of times.

    Parameters
    ----------
    samples : :class:`~thejoker.sampler.JokerSamples`
        Posterior samples from The Joker.
    t_grid : array_like, `~astropy.time.Time`, optional
        Array of times. Either in BMJD or as an Astropy Time object. If not
        specified, the time grid will be set to the data range with a small
        buffer.
    rv_unit : `~astropy.units.UnitBase`, optional
        The units to use when plotting RV's.
    data : `~thejoker.data.RVData`, optional
        Over-plot the data as well.
    ax : `~matplotlib.Axes`, optional
        A matplotlib axes object to plot on to. If not specified, will
        create a new figure and plot on that.
    plot_kwargs : dict, optional
        Passed to `matplotlib.pyplot.plot()`.
    data_plot_kwargs : dict, optional
        Passed to `thejoker.data.RVData.plot()`.
    add_labels : bool, optional
        Add labels to the axes or not.
    relative_to_t0 : bool, optional
        Plot the time axis relative to ``samples.t0``.

    Returns
    -------
    fig : `~matplotlib.Figure`

    """

    if ax is None:
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(1, 1)
    else:
        fig = ax.figure

    if t_grid is None:
        if data is None:
            raise ValueError('If data is not passed in, you must specify '
                             'the time grid.')

        w = np.ptp(data.t.mjd)
        dt = samples['P'].to(u.day).value.min() / 64
        t_grid = np.arange(data.t.mjd.min() - w*0.05,
                           data.t.mjd.max() + w*0.05 + dt,
                           dt)

        if len(t_grid) > 1e5:
            warnings.warn("Time grid has more than 100,000 grid points, so "
                          "plotting orbits could be very slow! Set 't_grid' "
                          "manually to decrease the number of grid points.",
                          ResourceWarning)

    if not isinstance(t_grid, Time): # Assume BMJD
        t_grid = Time(t_grid, format='mjd', scale='tcb')

    # scale the transparency of the lines
    n_plot = len(samples)
    Q = 4. # HACK
    line_alpha = 0.05 + Q / (n_plot + Q)

    if rv_unit is None:
        rv_unit = u.km/u.s

    # default plotting style
    style = plot_kwargs.copy()
    style.setdefault('linestyle', '-')
    style.setdefault('linewidth', 0.5)
    style.setdefault('alpha', line_alpha)
    style.setdefault('marker', '')
    style.setdefault('color', '#555555')
    style.setdefault('rasterized', True)

    # plot orbits over the data
    model_rv = np.zeros((n_plot, len(t_grid)))
    for i in range(n_plot):
        orbit = samples.get_orbit(i)
        model_rv[i] = orbit.radial_velocity(t_grid).to(rv_unit).value

    bmjd = t_grid.tcb.mjd
    if relative_to_t0:
        if samples.t0 is None:
            raise ValueError('Input samples object has no epoch .t0')
        bmjd = bmjd - samples.t0.tcb.mjd

    ax.plot(bmjd, model_rv.T, **style)

    if data is not None:
        data_style = data_plot_kwargs.copy()
        data_style.setdefault('rv_unit', rv_unit)
        data_style.setdefault('markersize', 4.)

        if data_style['rv_unit'] != rv_unit:
            raise u.UnitsError("Data plot units don't match rv_unit!")

        data.plot(ax=ax, relative_to_t0=relative_to_t0, add_labels=False,
                  **data_style)

        _rv = data.rv.to(rv_unit).value
        drv = _rv.max() - _rv.min()
        ax.set_ylim(_rv.min() - 0.2*drv, _rv.max() + 0.2*drv)

    ax.set_xlim(bmjd.min(), bmjd.max())
    if add_labels:
        ax.set_xlabel('BMJD')
        ax.set_ylabel('RV [{}]'
                      .format(rv_unit.to_string(format='latex_inline')))

    return fig



from ._astropy_init import *

# Enforce Python version check during package import.
# This is the same check as the one at the top of setup.py
import sys

__minimum_python_version__ = "3.6"

class UnsupportedPythonError(Exception):
    pass

if sys.version_info < tuple((int(val) for val in __minimum_python_version__.split('.'))):
    raise UnsupportedPythonError("packagename does not support Python < {}"
                                 .format(__minimum_python_version__))

# For egg_info test builds to pass, put package imports here.
if not _ASTROPY_SETUP_:
    from .sampler import *
    from .data import *
    from .utils import *



# Third-party
import astropy.units as u
import numpy as np
from scipy.stats import multivariate_normal
from twobody.units import UnitSystem

# Project
from .likelihood import get_ivar, design_matrix
from .params import JokerParams
from .samples import JokerSamples
from ..data import RVData
from ..stats import beta_logpdf, norm_logpdf

__all__ = ['TheJokerMCMCModel']

log_2pi = np.log(2 * np.pi)


def get_vterms(p, poly_trend):
    return dict([('v'+str(i), p['v'+str(i)])
                 for i in range(poly_trend)])


class TheJokerMCMCModel:

    def __init__(self, joker_params, data, units=None):
        """

        Parameters
        ----------
        data : `~thejoker.data.RVData`
            The radial velocity data.
        """

        # check if a JokerParams instance was passed in to specify the state
        if not isinstance(joker_params, JokerParams):
            raise TypeError("Parameter specification must be a JokerParams "
                            "instance, not a '{0}'".format(type(joker_params)))
        self.params = joker_params

        if not isinstance(data, RVData):
            raise TypeError("Data must be a valid RVData object.")
        self.data = data

        if units is None:
            units = UnitSystem(u.day, u.radian, u.Msun, u.au,
                               self.data.rv.unit)
        else:
            if not isinstance(units, UnitSystem):
                raise TypeError('Input units must be as a '
                                'twobody.units.UnitySystem object, not {}'
                                .format(units.__class__.__name__))
        self.units = units

        if self.params._fixed_jitter:
            s = self.params.jitter.to_value(self.units['speed'])
            self._ln_s2 = 2 * np.log(s)
            self._s = s
        else:
            self._ln_s2 = None
            self._s = None

        # Covariance matrix for Gaussian prior on velocity trend terms:
        #   K, v0, v1, v2 ...
        self._V = np.linalg.inv(self.params.linear_par_Vinv)
        self._vterms_norm = multivariate_normal(
            mean=np.zeros(1 + self.params.poly_trend),
            cov=self._V)

        # various cached things:
        self._P_min = self.params.P_min.to_value(self.units['time'])
        self._P_max = self.params.P_max.to_value(self.units['time'])
        self._rv = self.data.rv.value
        self._jitter_factor = self.units['speed'].to(self.params._jitter_unit)

    ###########################################################################
    # Probability expressions
    #
    def ln_likelihood(self, p):
        A = design_matrix([p['P'], p['M0'], p['e'], p['omega']],
                          self.data, self.params)
        vterms = np.ravel(list(get_vterms(p, self.params.poly_trend).values()))
        Kv_terms = np.concatenate((np.atleast_1d(p['K']), vterms))
        ivar = get_ivar(self.data, p['jitter'])
        dy = A.dot(Kv_terms) - self._rv

        return 0.5 * (-dy**2 * ivar - log_2pi + np.log(ivar))

    def ln_prior(self, p):
        # TODO: right now, the priors are hard-coded! These should be
        # customizable by the user

        lnp = 0.

        if not 0 < p['e'] < 1:
            return -np.inf
        lnp += beta_logpdf(p['e'], 0.867, 3.03) # Kipping et al. 2013

        # TODO: not normalized properly because we don't use a truncnorm for K
        if p['K'] < 0:
            return -np.inf

        # uniform in ln(P) - we don't need the jacobian because we sample in lnP
        # TODO: but this is not normalized correctly
        if not self._P_min < p['P'] < self._P_max:
            return -np.inf

        # TODO: priors on M0, omega not normalized properly (need 1/2pi?)

        if not self.params._fixed_jitter:
            # Gaussian prior in ln(s^2) - don't need Jacobian because we are
            # actually sampling in y = ln(s^2)
            s_scaled = p['jitter'] * self._jitter_factor
            y = 2 * np.log(s_scaled)
            lnp += norm_logpdf(y, self.params.jitter[0],
                               self.params.jitter[1])

        # Gaussian priors on K, v0, v1, ...
        # TODO: technicaly, K prior is improper because half-gaussian (K>0)!
        vterms = np.ravel(list(get_vterms(p, self.params.poly_trend).values()))
        Kv_terms = np.concatenate((p['K'], vterms))
        lnp += self._vterms_norm.logpdf(Kv_terms)

        return lnp

    def ln_posterior(self, mcmc_p):
        p = self.unpack_samples(mcmc_p, add_units=False)

        lnp = self.ln_prior(p)
        if not np.isfinite(lnp):
            return -np.inf

        lnl = self.ln_likelihood(p)
        lnprob = lnp + lnl.sum()

        if not np.isfinite(lnprob):
            return -np.inf

        return lnprob

    def __call__(self, mcmc_p):
        return self.ln_posterior(mcmc_p)

    ###########################################################################
    # Parameter transformations
    #

    def _strip_units(self, p):
        """Remove units from the input samples by converting to the standard
        unit system.
        """
        new_p = dict()

        new_p['P'] = p['P'].to_value(self.units['time'])
        new_p['M0'] = p['M0'].to_value(self.units['angle'])
        new_p['e'] = np.asarray(p['e'])
        new_p['omega'] = p['omega'].to_value(self.units['angle'])
        new_p['K'] = p['K'].to_value(self.units['speed'])

        if 'jitter' in p:
            new_p['jitter'] = p['jitter'].to_value(self.units['speed'])

        v_terms = get_vterms(p, self.params.poly_trend)
        for i, k in enumerate(v_terms.keys()):
            _unit = self.units['speed'] / self.units['time']**i
            new_p[k] = p['v'+str(i)].to_value(_unit)

        return new_p

    def _add_units(self, p):
        """Add units to the input samples from the standard unit system.
        """
        new_p = JokerSamples()

        new_p['P'] = p['P'] * self.units['time']
        new_p['M0'] = p['M0'] * self.units['angle']
        new_p['e'] = np.asarray(p['e'])
        new_p['omega'] = p['omega'] * self.units['angle']
        new_p['K'] = p['K'] * self.units['speed']

        if 'jitter' in p:
            new_p['jitter'] = p['jitter'] * self.units['speed']

        v_terms = get_vterms(p, self.params.poly_trend)
        for i, k in enumerate(v_terms.keys()):
            _unit = self.units['speed'] / self.units['time']**i
            new_p[k] = p['v'+str(i)] * _unit

        return new_p

    def pack_samples(self, p, strip_units=True):
        r"""Pack a dictionary of samples (in Keplerian parameters P, e, K, etc.)
        as Quantity objects into a 2D array of MCMC parameters.

        The standard variables for MCMC sampling are:

        .. math::

            \ln P \\
            \sqrt{K}\,\cos(M_0-\omega), \sqrt{K}\,\sin(M_0-\omega) \\
            \sqrt{e}\,\cos\omega, \sqrt{e}\,\sin\omega \\
            \ln s^2 \\
            v_0,..., v_n

        Parameters
        ----------
        arr : iterable
            A packed parameter vector containing the orbital parameters
            and long-term velocity trend parameters.

        """
        if strip_units:
            p = self._strip_units(p)

        terms = [np.log(p['P']),
                 np.sqrt(p['K']) * np.cos(p['M0'] - p['omega']),
                 np.sqrt(p['K']) * np.sin(p['M0'] - p['omega']),
                 np.sqrt(p['e']) * np.cos(p['omega']),
                 np.sqrt(p['e']) * np.sin(p['omega'])]

        if not self.params._fixed_jitter:
            terms.append(2 * np.log(p['jitter']))

        # Now add the velocity terms as linear parameters
        v_terms = get_vterms(p, self.params.poly_trend)
        terms.extend(list(v_terms.values()))

        return np.stack(terms)

    def unpack_samples(self, arr, add_units=True):
        """Unpack a 2D array of samples in standard variables form into a
        dictionary of samples as Quantity objects as Kepler parameters (i.e.
        period, angles, ...).

        Parameters
        ----------
        arr : `numpy.ndarray`
            A 2D numpy array with shape `(nsamples, ndim)` containing the
            sample values.

        Returns
        -------
        samples : dict
            Dictionary of `~astropy.units.Quantity` objects for period,
            M0, etc.
        """
        arr = np.atleast_2d(arr)

        (ln_P,
         sqrtK_cos_M0, sqrtK_sin_M0,
         sqrte_cos_omega, sqrte_sin_omega) = arr.T[:5]

        M0_minus_omega = np.arctan2(sqrtK_sin_M0, sqrtK_cos_M0)
        omega = np.arctan2(sqrte_sin_omega, sqrte_cos_omega)
        M0 = (M0_minus_omega + omega)

        M0 = M0 % (2*np.pi)
        omega = omega % (2*np.pi)
        e = sqrte_cos_omega**2 + sqrte_sin_omega**2
        K = sqrtK_cos_M0**2 + sqrtK_sin_M0**2

        # Load the Keplerian sample values:
        samples = dict()

        samples['P'] = np.exp(ln_P)
        samples['M0'] = M0
        samples['e'] = e * u.one
        samples['omega'] = omega
        samples['K'] = K

        if not self.params._fixed_jitter:
            samples['jitter'] = arr.T[5]
            k = 1
        else:
            samples['jitter'] = np.full(len(arr), self._s)
            k = 0

        for i in range(self.params.poly_trend):
            samples['v' + str(i)] = arr.T[5 + k + i]

        if add_units:
            samples = self._add_units(samples)

        return samples



# Standard library
import logging

# Third-party
from astropy.extern.six import PY3
from astropy.logger import StreamHandler
from astropy.utils import find_current_module

__all__ = ['log']


Logger = logging.getLoggerClass()
class JokerLogger(Logger):

    def _set_defaults(self):
        """
        Reset logger to its initial state
        """

        # Remove all previous handlers
        for handler in self.handlers[:]:
            self.removeHandler(handler)

        # Set levels
        self.setLevel(logging.INFO)

        # Set up the stdout handler
        sh = StreamHandler()
        self.addHandler(sh)

    def makeRecord(self, name, level, pathname, lineno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        if extra is None:
            extra = {}
        if 'origin' not in extra:
            current_module = find_current_module(1, finddiff=[True, 'logging'])
            if current_module is not None:
                extra['origin'] = current_module.__name__
            else:
                extra['origin'] = 'unknown'

        if PY3:
            return Logger.makeRecord(self, name, level, pathname, lineno, msg,
                                     args, exc_info, func=func, extra=extra,
                                     sinfo=sinfo)
        else:
            return Logger.makeRecord(self, name, level, pathname, lineno, msg,
                                     args, exc_info, func=func, extra=extra)


logging.setLoggerClass(JokerLogger)
log = logging.getLogger('thejoker')
log._set_defaults()



# Standard library
from os import path

# Third-party
import astropy.units as u
from astropy.tests.helper import quantity_allclose
import numpy as np
import pytest

# Package
from ..params import JokerParams
from ..sampler import TheJoker
from .helpers import FakeData


class TestSampler(object):

    def truths_to_nlp(self, truths):
        P = truths['P'].to(u.day).value
        M0 = truths['M0'].to(u.radian).value
        ecc = truths['e']
        omega = truths['omega'].to(u.radian).value
        return np.array([P, M0, ecc, omega, 0.])

    def setup(self):
        d = FakeData()
        self.data = d.datasets
        self.joker_params = d.params
        self.truths = d.truths

    def test_init(self):
        TheJoker(self.joker_params['binary'])
        # TheJoker(self.joker_params['triple']) # TODO: disabled

        # invalid pool
        class DummyPool(object):
            pass

        pool = DummyPool()
        with pytest.raises(TypeError):
            TheJoker(self.joker_params['binary'], pool=pool)

        # invalid random state
        rnd = "aisdfijs"
        with pytest.raises(TypeError):
            TheJoker(self.joker_params['binary'], random_state=rnd)

        # invalid params
        pars = "aisdfijs"
        with pytest.raises(TypeError):
            TheJoker(pars)

        j = TheJoker(self.joker_params['binary'],
                     tempfile_path=path.abspath('.'))

    def test_sample_prior(self):
        rnd = np.random.RandomState(42)
        joker = TheJoker(self.joker_params['binary'], random_state=rnd)
        samples = joker.sample_prior(8)
        samples, ln_vals = joker.sample_prior(8, return_logprobs=True)
        assert np.isfinite(ln_vals).all()

    def test_rejection_sample(self):
        rnd = np.random.RandomState(42)

        # First, try just running rejection_sample()
        data = self.data['binary']
        joker = TheJoker(self.joker_params['binary'], random_state=rnd)

        with pytest.raises(ValueError):
            joker.rejection_sample(data)

        joker.rejection_sample(data, n_prior_samples=128)

        # Now try specifying the tempfile path
        data = self.data['binary']
        joker = TheJoker(self.joker_params['binary'],
                         random_state=rnd, tempfile_path=path.abspath('.'))

        _ = joker.rejection_sample(data, n_prior_samples=128)

        # Now re-run with jitter set, check that it's always the fixed value
        jitter = 5.*u.m/u.s
        params = JokerParams(P_min=8*u.day, P_max=128*u.day, jitter=jitter)
        joker = TheJoker(params)

        prior_samples = joker.sample_prior(128)
        assert quantity_allclose(prior_samples['jitter'], jitter)

        full_samples = joker.rejection_sample(data, n_prior_samples=128)
        assert quantity_allclose(full_samples['jitter'], jitter)

        samples, lnp, lnl = joker.rejection_sample(data,
                                                   n_prior_samples=128,
                                                   return_logprobs=True)
        assert len(lnp) == len(samples)
        assert len(lnl) == len(samples)

    def test_iterative_rejection_sample(self):

        # First, try just running rejection_sample()
        data = self.data['binary']
        jitter = 100*u.m/u.s
        params = JokerParams(P_min=8*u.day, P_max=128*u.day, jitter=jitter)
        joker = TheJoker(params)

        samples = joker.iterative_rejection_sample(data, n_prior_samples=100000,
                                                   n_requested_samples=2)

        assert quantity_allclose(samples['jitter'], jitter)

        samples, lnp, lnl = joker.iterative_rejection_sample(
            data, n_prior_samples=100000, n_requested_samples=2,
            return_logprobs=True)
        assert len(lnp) == len(samples)
        assert len(lnl) == len(samples)

    def test_mcmc_continue(self):
        rnd = np.random.RandomState(42)

        # First, try just running rejection_sample()
        data = self.data['binary']
        joker = TheJoker(self.joker_params['binary'], random_state=rnd)

        samples = joker.rejection_sample(data, n_prior_samples=16384)
        joker.mcmc_sample(data, samples, n_steps=8, n_burn=8,
                          n_walkers=128, return_sampler=False)
        joker.mcmc_sample(data, samples, n_steps=8, n_burn=8, n_walkers=128,
                          return_sampler=True)



# Third-party
from astropy.tests.helper import quantity_allclose
import astropy.units as u
import numpy as np

# Package
from ..mcmc import TheJokerMCMCModel
from ..params import JokerParams
from .helpers import FakeData


class TestMCMC(object):

    def truths_to_nlp(self, truths):
        # P, M0, ecc, omega
        P = truths['P'].to(u.day).value
        M0 = truths['M0'].to(u.radian).value
        ecc = truths['e']
        omega = truths['omega'].to(u.radian).value
        return np.array([P, M0, ecc, omega, 0.])

    def setup(self):
        d = FakeData()
        self.data = d.datasets
        self.joker_params = d.params
        self.truths = d.truths

    def test_funcs(self):
        data = self.data['binary']
        truth = self.truths['binary']
        params = self.joker_params['binary']
        model = TheJokerMCMCModel(params, data)

        # test roundtrip
        mcmc_p = model.pack_samples(truth)
        p2 = model.unpack_samples(mcmc_p)
        for k in truth:
            assert quantity_allclose(p2[k], truth[k])

        assert 'jitter' in p2 and quantity_allclose(p2['jitter'],
                                                    params.jitter)

        lp = model.ln_prior(model._strip_units(p2))
        assert np.isfinite(lp)

        ll = model.ln_likelihood(model._strip_units(p2))
        assert np.isfinite(ll).all()

        lnpost = model(mcmc_p)
        assert np.isfinite(lnpost)
        assert np.allclose(lnpost, lp+ll.sum())
